; +--------------------------------------------------+
; | NexGen Flash BIOS update utility PCI-Bus         |
; +--------------------------------------------------+

		title "NexGen Flash BIOS update utility PCI-Bus"

; versions: (keep in sync with make.bat)
PCI10		equ vers EQ 10
PCI30		equ vers EQ 30
PCI31		equ vers EQ 31

		.nolist
include biosdata.inc
include screen.inc
		.listall

StackSize	equ 200h
ImageAddress	equ 4000h		; in paragraphs
BiosAddress	equ 0e000h		; in paragraphs

EditBufSize	equ 40h

; ASCII character codes
BS		equ 8
LF		equ 10
CR		equ 13
ESCP		equ 27
DEL		equ 127

; Flash EPROM commands
FE_EraChip	equ 010h		; erase chip
FE_DPDisable	equ 020h		; software data protect disable
FE_EraSector	equ 030h		; erase sector
FE_Alternate	equ 060h		; alternate software product ID
FE_Prefix	equ 080h		; for 6 cycle commands
FE_Autoselect	equ 090h		; read device ID
FE_Program	equ 0A0h		; program byte(s)
FE_EraSecSusp	equ 0B0h		; during sector erase: suspend
					; resume via FE_EraSector
FE_Reset	equ 0F0h		; read/reset

; usage for Advanced Micro Devices (AMD) 29F010A:

;                            <- 1 -> <- 2 -> <- 3 -> <- 4 -> <- 5 -> <- 6 ->
;                        Cyc Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta
; Read (4)               1   RA  RD
; Reset (5)              1   xxx F0
; Reset (6)              3   555 AA  2AA 55  555 F0
; Autoselect (7):
; Manufacturer ID        4   555 AA  2AA 55  555 90  X00 01
; Device ID              4   555 AA  2AA 55  555 90  X01 20
; Sector Prot.Verify (8) 4   555 AA  2AA 55  555 90  SA+ 00/
;                                                    X02 01
; Program                4   555 AA  2AA 55  555 A0  PA  PD
; Chip Erase             6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 10
; Sector Erase           6   555 AA  2AA 55  555 80  555 AA  2AA 55  SA  30
; Erase Suspend (9)      1   XXX B0
; Erase Resume (10)      1   XXX 30

; usage for Silicon Storage Technology (SST) SST29EE010:

;                            <- 1 -> <- 2 -> <- 3 -> <- 4 -> <- 5 -> <- 6 ->
;                        Cyc Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta
; Software Data
; Protect Enable
; & Page Write         3+128 555 AA  2AA 55  555 A0  PA  PD ...
; Software Data
; Protect Disable        6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 20
; Software Chip Erase    6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 10
; Software ID Entry      3   555 AA  2AA 55  555 90
; Software ID Exit       3   555 AA  2AA 55  555 F0
; Alternate Software
; ID Entry               6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 60


; useful macros
; define zero terminated string
dz		macro	n:req
		db n, 0
		endm

; combine x and y coordinates into word
yx		macro	y:req, x:req
		exitm	%x+(y*256)
		endm

; special NexGen opcode to control various undocumented processor features
nxop		macro
		db 0fh, 55h
		endm


flashp		segment public 'CODE'
		assume cs:flashp
		assume ds:nothing, es:nothing, ss:nothing

; +--------------------------------------------------+
; | VARIABLES and CONSTANTS                          |
; +--------------------------------------------------+

; program segment prefix (PSP)
		db 80h dup(?)		; unused
CmdLine		db 80h dup(?)

; program entry point
Start:		jmp	Start03
		cli			; unused. Maybe align/even?

; v-- patch area for screen codes --v
TxUppLeftCornr:	dz ' '
TxUppRghtCornr:	dz ' '
TxLwrLeftCornr:	dz ' '
TxLwrRghtCornr:	dz ' '
TxHorizLine:	dz ' '
TxVertLine:	dz ' '
		dz ' '			; unused
		dz ' '			; unused
; ^-- patch area for screen codes --^

TxSpaceLine:	db 84 dup(' ')
		db 0

TxSaveFilename:	db EditBufSize dup(0)
		db 100h dup(0)		; unused

Tx37NewLine: rept 37
		db LF, CR
	     endm
		db 0

StackArea:	dw StackSize dup(0)

		db 200h dup(0)		; unused
		dw 0
DelayFact10	dw 0			; loops per 10 usec
DelayFact6	dw 0			; loops per  6 usec
		dw 0

TxWriteFileErr: dz <CR,'write file error',CR,LF>
TxCreateErr:	dz <CR,'Create error',CR,LF>

TxSignon:	dz 'NEXGEN SYSTEM ROM BIOS UPDATE UTILITY'

TxDate:
 IF PCI10
		dz '10/20/95'
 ELSE
		dz '01/19/96'
 ENDIF

TxEnterUpdate:	dz 'Enter "FILENAME.XXX" to update in ROM, then press ENTER key.'

TxEnterSave:	dz 'Enter FILENAME.BIN to save modified BIOS, then press ENTER key.'

TxFileNotExist:	dz 'FILE DOES NOT EXIST, press any key to continue.'

TxFileSizeErr:	dz 'File size read error, press any key to continue.'

TxCantCreate:	dz 'Cannot create Filename, press any key to continue.'

TxPleaseWait:
 IF PCI10
		db '       '
 ENDIF
		dz 'Please wait.  Reading the BIOS update information ...'

TxNotEnoughMem:	dz 'CANNOT CONTINUE,.. NOT ENOUGH LOW MEMORY AVAILABLE.'

TxContAbort:
 IF PCI10
		db '            '
 ENDIF
		db 'PRESS "C" TO CONTINUE'
 IF PCI10
		dz 0
 ELSE
		dz ' OR "ESC" TO ABORT'
 ENDIF

TxFileWriteErr:	dz 'File write error, press any key to continue.'

TxChecksumErr:
 IF PCI10
		dz 'FILE HAVE BIOS CHECKSUM ERROR, Press any key to continue'
 ELSE
		dz 'FILE HAS BIOS CHECKSUM ERROR, Press any key to continue'
 ENDIF

TxDateIdentic:	db 'CANNOT CONTINUE, FILE AND ROM DATA ARE IDENTICAL, '
		dz 'press a key to restart.'

TxNoNexGenBios:	dz 'FILE IS NOT A VALID NEXGEN BIOS, Press any key to continue'

TxWrongSystem:	dz 'YOU CANNOT USE THIS UTILITY ON THIS SYSTEM, Press any key to continue'

TxUpdComplete:	dz 'BIOS UPDATE COMPLETE'

TxWriteProtect:	dz 'Cannot continue...., FLASH MEMORY is write protected.'

TxBadFlashRom:	dz 'Cannot continue...., unidentified or bad FLASH ROM.'

TsPleaseNote:	dz 'PLEASE NOTE:'
		dz ' '
		dz '----------- '
		dz ' '
		dz 'THIS UTILITY PROGRAM MAY BE AFFECTED BY OTHER DRIVERS LOADED'
		dz 'IN MEMORY.  IF YOU ENCOUNTER ANY PROBLEMS DURING THE UPDATE'
		dz 'PROCESS, YOU SHOULD RESET THE SYSTEM AND DO A CLEAN BOOT.'
		dz '(See the README.TXT for directions on how to do a clean boot.'
		dz 'If you want to to read the file NOW, press "ESCAPE" key to abort'
		dz 'this program.  Then, enter "TYPE README.TXT|MORE")'
		dz ' '
		dz 'AFTER THE CLEAN BOOT, RUN THE UPDATE UTILITY AGAIN.'
		dz ' '
 IF PCI10
		dz '               Press "C" to continue ...'
 ELSE
		dz 'Press "C" to continue or "ESC" to abort ...'
		dz ' '
		dz '(This info is also available on NexGen Online at '
		dz ' http://www.nexgen.com)'
 ENDIF
		db 0

TsThankYou:
 IF PCI10
		db '               '
 ENDIF
		dz 'Thank you for choosing NexGen.'
		dz ' '
		dz 'You have activated the NexGen BIOS flash utility for your'
		dz 'Nx586 PCI-based PC.  This utility will update your BIOS'
		dz 'quickly and effectively.'
		dz ' '
 IF PCI10
		db '               '
 ENDIF
		dz 'Press any key to continue ...'
		db 0

TxDot:		dz '.'

TxDefaultDrive:	dz 'Default Drive Directory:'

TsOptions:	dz ' option available:'
		dz '       /v = will erase NVRAM'
		dz '       /fFILENAME.EXT = File to flash'
		dz '       /q = quick flash'
		dz '       /s = will save current BIOS to a file (filename=BIOS.OLD)'
		db 0

TsFatalError:	dz '                  F A T A L  E R R O R'
		dz ' '
 IF PCI10
		db '  '
 ENDIF
		dz 'CANNOT UPDATE ROM BIOS, FLASH ROM WRITE ERROR DETECTED'
		dz 'SYSTEM WILL NOT FUNCTION ONCE YOU RESET THE SYSTEM, CONTACT'
		dz 'YOUR DISTRIBUTOR FOR ROM BIOS REPLACEMENT.'
		db 0

TsSuccess:	dz ' '
 IF PCI10
		db '       '
 ENDIF
		dz 'System BIOS has been succesfully upgraded!'
		dz ' '
 IF PCI10
		db '  '
 ENDIF
		dz 'TURN OFF THE COMPUTER, THEN BACK ON, TO ACTIVATE NEW BIOS'
		db 0

TxBiosVersion:	dz 'BIOS VERSION TO FLASH IS              '

TsFinalWarning:
 IF PCI10
		db '             '
 ENDIF
		dz 'FINAL WARNING !!!'
		dz '                                           '
		dz 'THIS IS YOUR LAST CHANCE TO DISCONTINUE UPDATE,'
		dz 'YOU CAN PRESS ESC NOW TO ABORT UPDATE.'
		dz 'IF YOU SELECT TO CONTINUE, UPDATING MAY TAKE UP'
		dz 'TO 30 SECONDS. DO NOT RESET OR POWER DOWN THE'
		dz 'SYSTEM WHILE UPDATING.'
		db 0

TxAbortUpdate:	dz 'ABORT BIOS UPDATE.'

TxBiosVersion2:	dz '                                           '

TxBeingUpdated:	dz 'The BIOS is being updated ... '

TsDoNotReset:	dz 'DO NOT RESET OR TURN-OFF SYSTEM.'
		dz 'UPDATING MY TAKE UP TO 30 SECONDS.'
		dz 'WAIT FOR COMPLETION MESSAGE.'
		db 0


FileHandle1	dw 0
FileHandle2	dw 0
CharAttr	db BG_Blue+FG_White
		db BG_Blue+FG_White

TsCursorKeys:	dz 'Insert key -> to insert a character      Left cursor key --> to go left'
		dz 'Delete key -> to delete a character      Right cursor key -> to go right'
		dz '                              ESC --> To abort'
		db 0

TxControlKeys:	dz 'Control Keys'

 IF PCI10
TxBiosFileMC3:	db 'P06W034N.MC3'
ProgRetries	db 0
 ELSEIF PCI30
TxBiosFilename:	dz 'P06Y034P.PRD'
 ELSE ;PCI31
TxBiosFilename:	dz 'P06Y034Q.PRD'
 ENDIF

 IF not PCI10
TsVlBiosOnPci:	dz 'You are attempting to update a VL system BIOS with a PCI'
		dz 'update utility.  You must use the VL version.  This utility'
		dz 'is part of the update kit: UPDTVLxx.EXE (where "xx"'
		dz 'is the version number).  It can be obtained from NexGen Online:'
		dz '               BBS - 408.955.1839'
		dz '               WEB - http://www.nexgen.com'
		dz ' '
		dz 'Press any key ...'
		db 0

ProgRetries	db 0
 ENDIF

ProgTimeout	db 0
OptionSV:
OptionV		db 0
OptionS		db 0
OptionQF:
OptionF		db 0
OptionQ		db 0

TxMagic1:	db 'angel'
TxMagic2:	db '@(#)0@(#)x'

FileNameBuffer:	db 45 dup(0)

 IF PCI10
TxBiosFilename:	db 'P06w034n.prd'
 ELSE
		db 'P06Y034P.PRD'

TsLatestVers:	dz 'The BIOS on your system is the latest version.'
		dz 'No update is necessary.'
		dz ' '
		dz 'Press any key ...'
		db 0
 ENDIF

ImageSegment	dw ImageAddress


; +--------------------------------------------------+
; | CODE                                             |
; +--------------------------------------------------+

; Scan command line of PSP
; E: DS = ES = PSP, CS = SS = code segment

Start03:	cld
		cli
		mov	bx,offset CmdLine
		xor	cx,cx
		mov	cl,[bx]		; get length of argument string
		or	cl,cl		; command line empty?
		jnz	Start06		; no, process line
		jmp	Main

Start06:	; process command line
		inc	bx
		mov	al,[bx]		; get char
		cmp	al,'?'		; user wants help?
		jnz	Start09		; no, check for options
		jmp	short Start33	; yes, abort scanning

Start09:	and	al,0DFh		; convert to upper case
		cmp	al,'V'		; erase NVRAM?
		jnz	Start12		; no, check for next option
		mov	cs:OptionV,'V'	; yes, flag it
		loop	Start06		; next char available?
		jmp	short Start30	; no, done with scan

Start12:	cmp	al,'S'		; save current BIOS?
		jz	Start24		; yes, flag it
		cmp	al,'Q'		; quick flash?
		jnz	Start15		; no, check for next option
		mov	cs:OptionQ,'Q'	; yes, flag it
		loop	Start06		; next char available?
		jmp	short Start30	; no, done with scan

Start15:	cmp	al,'F'		; file to flash?
		jnz	Start27		; no, unknown option, ignore
		inc	bx
		mov	al,[bx]		; skip separators
		cmp	al,' '		; space?
		jz	Start27		; yes, ignore
		cmp	al,CR		; end of command line?
		jz	Start27		; yes, ignore
		cmp	al,'/'		; new option?
		jz	Start27		; yes, ignore
		dec	cx		; was this the last char?
		jz	Start30		; yes, ignore
		inc	cx
		dec	bx
		mov	ax,cs		; now copy filename from
		mov	es,ax		; command line into buffer
		cld
 IF PCI10
		call	GetBiosFName
 ELSE
		mov	di,offset TxBiosFilename
 ENDIF

Start18:	inc	bx
		mov	al,[bx]		; until delimiter encountered
		cmp	al,' '		; space?
		jz	Start21		; yes, end of filename
		cmp	al,'/'		; new option?
		jz	Start21		; yes, end of filename
		cmp	al,CR		; end of command line?
		jz	Start21		; yes, end of filename
		stosb			; no, save valid filename char
		loop	Start18		; continue until end of command line
		mov	cs:OptionF,'F'	; flag option
		jmp	short Start30	; done with scan

Start21:	; end of filename reached, but command line not exhausted
		mov	cs:OptionF,'F'	; flag option
		jmp	short Start27	; and look out for more options

Start24:	; set flag "save current BIOS"
		mov	cs:OptionS,'S'	; flag option

Start27:	; goto next char (if available)
		loop	Start06

Start30:	; end of command line reached
		cmp	word ptr cs:OptionSV,0	; any option set?
		jnz	Main		; yes
		cmp	word ptr cs:OptionQF,0
		jz	Start33		; no, help user
		jmp	short Main	; yes

Start33:	; user help
		mov	cs:OptionV,0FFh	; flag help
		jmp	short Main	; unnecessary

Main:		; OptionV = 0FFh for help, otherwise OptionX = X if option set
		; TxBiosFilename contains name if OptionF set
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	ss,ax
		mov	sp,offset StackArea+2*StackSize
 IF not PCI10
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_White+FG_Black	; no effect
		call	SetCharAttr		;  "
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	si,0B000h	; my NX586 PCI returns:
		nxop			; AH=00, BX=0004, CX=1504, DX=3330
		cmp	cl,6		; check BIOS version. Current?
		jnz	Main06		; no

		; no update necessary - tell user and exit
		mov	ax,yx(3,7)
		mov	bx,offset TsLatestVers
		call	PrintPosMulti
		call	GetKeyb
		jmp	TerminateProg

Main03:		; Nx586 VL system - tell user and exit
		mov	ax,yx(3,7)
		mov	bx,offset TsVlBiosOnPci
		call	PrintPosMulti
		call	GetKeyb
		jmp	TerminateProg

Main06:		; BIOS version in CL
		cmp	cl,4		; BIOS of NX586 PCI system?
		jnz	Main03		; no, must be VL system
 ENDIF
		mov	si,offset Tx37NewLine
		call	Print		; a simple form of ClearScreen
		cmp	cs:OptionV,0FFh	; user help?
		jnz	Main09		; no, continue

		; show user help screen (option syntax) and exit
		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(7,0)
		mov	bx,offset TsOptions
		call	PrintPosMulti
		jmp	TerminateProg

Main09:		; no help, check memory requirements
		mov	ax,cs		; check if end of program below image address
		cmp	ax,ImageAddress-500h ; reserve 20KB for program
		jbe	Main12

		; not enough memory - tell user and exit
		mov	bx,offset TxNotEnoughMem
		mov	ax,yx(8,7)
		call	PrintPos
		jmp	PrintTermProg

Main12:		mov	si,0B001h	; switch off caches (1st and 2nd)
		mov	al,0
		nxop
		cmp	cs:OptionS,'S'	; save current BIOS option set?
		jnz	Main18		; no

		; save current BIOS
		; copy from Flash ROM into RAM (128 KB)
		cli			; needed: prevent Interrupts into BIOS because
		mov	al,0		; BIOS shadow RAM image is swapped now with BIOS EPROM image
		call	near ptr ProgChipSet
		mov	ax,cs:ImageSegment
		cld
		mov	es,ax
		mov	ax,BiosAddress
		mov	ds,ax
		mov	cx,8000h	; copy 32K words
		xor	di,di
		xor	si,si
	    rep movsw			; move CX words from DS:SI to ES:DI
		mov	ax,es		; increment segment
		add	ax,1000h
		mov	es,ax
		mov	ax,BiosAddress+1000h
		mov	ds,ax
		mov	cx,8000h	; copy 32 K words
		xor	di,di
		xor	si,si
	    rep movsw			; move CX words from DS:SI to ES:DI
		mov	al,1
		call	near ptr ProgChipSet
		sti

		; then save into file
		call	SaveOldBios	; save current BIOS into file. error?
		jb	Main15		; yes, then exit program
		mov	bx,cs:FileHandle1
		call	CloseFile
		jmp	short Main18

Main15:		; error - exit
		jmp	TerminateProg

Main18:		; clear image area
		cld
		push	es
		mov	ax,cs:ImageSegment
		mov	es,ax
		xor	di,di
		xor	ax,ax
		mov	cx,8000h	; fill 32K words with 0
	    rep stosw			; fill CX times AX to ES:DI
		mov	ax,es
		add	ax,1000h
		mov	es,ax
		xor	di,di
		xor	ax,ax
		mov	cx,8000h	; fill 32K words with 0
	    rep stosw			; fill CX times AX to ES:DI
		pop	es

		; show first screen
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_White+FG_Black ; no effect
		call	SetCharAttr	     ;  "
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	ax,yx(3,4)
		mov	bx,offset TsThankYou
		call	PrintPosMulti
		call	GetKeyb

		; show second screen
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_White+FG_Black ; no effect
		call	SetCharAttr	     ;  "
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	ax,yx(3,7)
		mov	bx,offset TsPleaseNote
		call	PrintPosMulti
		call	ContinueAbort

 IF not PCI10
		; show third screen
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_White+FG_Black ; no effect
		call	SetCharAttr	     ;  "
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
 ENDIF
		jmp	Main27

	; unused code >>
		; remove trailing blanks from TxBiosFilename
		.386
 IF PCI10
mc3		macro	op1:req, op2
		local ismc3, done
		mov	eax,80000000h	; if macro is put elsewhere,
		mov	dx,0FCF8h	; MASM reverses byte order !?!?!?
		out	dx,eax
		mov	dl,0FCh
		in	eax,dx
		clc
		cmp	al,3
		jnb	ismc3
		stc
OBiosFilename	textequ	<offset TxBiosFilename>
		op1
		op2
		jmp	short done

OBiosFilename	textequ	<offset TxBiosFileMC3>
ismc3:		op1
		op2
done:
		endm

		push	eax
		push	dx
		mc3	<mov bx,OBiosFilename>
		stc
		pop	dx
		pop	eax
 ELSE
		mov	bx,offset TxBiosFilename
		stc
 ENDIF

Main21:		cmp	byte ptr [bx],0 ; advance to terminating 0
		jz	Main24
		inc	bx
		jmp	short Main21

Main24:		dec	bx		; check char before 0
		cmp	byte ptr [bx],' '
		jz	Main24
		inc	bx
		mov	byte ptr [bx],0	; new end of string
		; enter editor
 IF PCI10
		push	bx
		mc3	<mov bx,OBiosFilename>
 ELSE
		stc
		mov	bx,offset TxBiosFilename
 ENDIF
		mov	ax,yx(4,7)
		call	PrintPos
 IF PCI10
		pop	bx
		push	eax
		push	dx
		mc3	<mov si,OBiosFilename-1>,<mov di,OBiosFilename-1+EditBufSize>
		stc
		pop	dx
		pop	eax
 ELSE
		stc
		mov	si,offset TxBiosFilename-1
		mov	di,offset TxBiosFilename-1+EditBufSize
		stc
 ENDIF
		call	Editor
	; << unused code

Main27:		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_White+FG_Black ; no effect
		call	SetCharAttr	     ;  "
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
 IF PCI10
		mc3	<xor dx,dx>,<mov dx,OBiosFilename>
 ELSE
		stc
		xor	dx,dx		; no effect
		mov	dx,offset TxBiosFilename
 ENDIF
		call	OpenFile
		jnb	Main30		; no error
		jmp	EFileNotExist	; open failed

Main30:		; file is open, filehandle in AX
		; read BIOS file (128K) into memory
		mov	cs:FileHandle1,ax
		push	ax
		mov	bx,offset TxPleaseWait
		mov	ax,yx(9,11)
		call	PrintPos
		pop	ax
		mov	bx,ax
		xor	dx,dx
		call	Read32KFromFile
		push	ax
		mov	ax,yx(11,30)
		call	PrintDot
		pop	ax
		mov	bx,cs:FileHandle1
		mov	dx,8000h
		call	Read32KFromFile
		push	ax		; not necessary
		mov	ax,yx(11,32)
		call	PrintDot
		pop	ax
		mov	ax,cs:ImageSegment
		add	ax,1000h
		mov	cs:ImageSegment,ax
		xor	dx,dx
		call	Read32KFromFile
		push	ax
		mov	ax,yx(11,34)
		call	PrintDot
		pop	ax
		mov	bx,cs:FileHandle1
		mov	dx,8000h
		call	Read32KFromFile
		push	ax		; not necessary
		mov	ax,yx(11,36)
		call	PrintDot
		pop	ax
		mov	ax,cs:ImageSegment
		sub	ax,1000h
		mov	cs:ImageSegment,ax

		; Check for magic 1234h at 4012h (corresponds to address 0E4012h).
		; This works for VL-Bios only, PCI-Bios has magic at 0E0012h.
		; Result is discard .. What a crap!
		push	ds
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,ds:4012h
		pop	ds
		cmp	ax,1234h	; some magic number
		jz	Main33		; no effect

Main33:		; Check Checksum. Result is discarded .. What a crap!
		; Read word at offset 0FFFAh (corresponds to address 0EFFFAh)
		push	ds
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,ds:0FFFAh
		pop	ds
		cmp	ax,0		; no checksum?
		jz	Main36		; then ignore it
		; checksum present, check it
		call	CheckChecksum
		jnb	Main36		; no effect

Main36:		; Compare BIOS image with current BIOS
		; Another nonsense: this compare will always fail
		; because, to support BIOS calls, the shadow RAM
		; image is currently mapped, not the EPROM image
		cli
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,0E000h
		mov	es,ax
		mov	cx,8000h	; 32K words
		xor	si,si
		xor	di,di
Main39:		lodsw
		scasw
		jnz	Main45
		loop	Main39
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
		mov	ax,0F000h
		mov	es,ax
		mov	cx,8000h
		xor	si,si
		xor	di,di
Main42:		lodsw
		scasw
		jnz	Main45
		loop	Main42
		; BIOS identical
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		sti
		mov	bx,offset TxDateIdentic
		jmp	EClosePrTerm	; close file, print error message,
					; then terminate program

Main45:		; BIOS different
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		sti
		mov	bx,cs:FileHandle1
		call	CloseFile

		; calibrate delay loops
		; uses BIOS data timer_low, which is incremented every
		; 55 ms (18.2 x per second) by the timer interrupt routine
		push	ds
		mov	ax,bdata
		mov	ds,ax
		assume ds:bdata
		mov	si,1000h
Calibr03:	mov	dl,16
		mov	cx,0
		; synchronise with timer tick for accurate measurement
		mov	ax,timer_low
Calibr06:	cmp	timer_low,ax
		jz	Calibr06
		; now time 16 x 64K loops
		mov	ax,timer_low
Calibr09:	loop	Calibr09	; loop 64K times
		dec	dl
		jnz	Calibr09
		mov	bx,timer_low
		cmp	bx,ax		; timer wrap around?
		jbe	Calibr03	; yes, repeat measurement
		sub	bx,ax		; BX = ticks (55 ms) per 16*64K loops
		mov	ax,si
		xor	dx,dx
		div	bx		; divide 1000h by ticks
		xor	dx,dx
		mov	bx,55		; msec per tick
		div	bx
		; AX = loops per 1/256 msec
		; Shift AX left 0, 4 or 8 before division by 100 for
		; maximum accuracy, then shift left again 8, 4 or 0
		; i.e. multiply by 256
		mov	cx,0800h
		cmp	ax,1000h	; faster than 1000h loops per 1/256 msec?
		jnb	Calibr15	; yes
		mov	cx,0404h
		cmp	ax,0100h	; faster than 100h loops per 1/256 msec?
		jnb	Calibr12	; yes
		mov	cx,0008h	; no, slower
Calibr12:	shl	ax,cl		; shift by 8, 4 or not
Calibr15:	push	ax
		xor	dx,dx
		mov	bx,100
		div	bx
		mov	cl,ch
		or	cl,cl
		jz	Calibr18
		shl	ax,cl
Calibr18:	; AX = loops per 1/100 msec (10 usec)
		mov	cs:DelayFact10,ax
		pop	ax
		xor	dx,dx
		mov	bx,167
		div	bx
		mov	cl,ch
		or	cl,cl
		jz	Calibr21
		shl	ax,cl
Calibr21:	; AX = loops per 1/167 msec (6 usec)
		mov	cs:DelayFact6,ax
		pop	ds

		; do option 'Q'
		push	ax
		push	bx
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		pop	bx
		pop	ax
 IF PCI10
		mov	ax,yx(15,13)
		mov	bx,offset TxContAbort
		call	PrintPos
 ENDIF
		mov	al,BG_Blue+FG_LtRed
		call	SetCharAttr
		cmp	cs:OptionQ,'Q'	; quick flash?
		jnz	ShowVers
		jmp	FlashProg	; yes, no final question

ShowVers:	; no quick flash - show version to flash
		; and ask user again if he is sure
		; search for BIOS version string
		mov	ax,cs
		mov	es,ax
		mov	ax,cs:ImageSegment
		add	ax,1000h
		mov	ds,ax
		mov	si,8000h
		mov	di,offset TxMagic1
		; ES:DI = ptr to TxMagic1
		; DS:SI = ptr into loaded BIOS image equivalent to 0F8000h
ShowVers03:	push	si
		push	di
		mov	cx,5
		cld
	    rep cmpsb
		jz	ShowVers06
		pop	di
		pop	si
		inc	si
		cmp	si,0FFFFh
		jnz	ShowVers03
		jmp	short ShowVers27

ShowVers06:	pop	di
		pop	di
		mov	di,offset TxBiosVersion+19h
ShowVers09:	lodsb
		cmp	al,0
		jz	ShowVers12
		stosb
		jmp	short ShowVers09

ShowVers12:	push	di
		mov	ax,ds
		sub	ax,1000h
		mov	ds,ax
		xor	si,si
		mov	di,offset TxMagic2
ShowVers15:	push	si
		push	di
		mov	cx,4
		cld
	    rep cmpsb
		jz	ShowVers21

ShowVers18:	pop	di
		pop	si
		inc	si
		cmp	si,8000h
		jnz	ShowVers15
		pop	di
		jmp	short ShowVers27

ShowVers21:	lodsb
		cmp	al,'0'
		jz	ShowVers24
		cmp	al,'x'
		jz	ShowVers24
		jmp	short ShowVers18

ShowVers24:	pop	di
		pop	di
		pop	di
		stosb
		mov	cx,3
	    rep movsb
		mov	ax,cs
		mov	ds,ax
		mov	si,offset TxBiosVersion
		mov	di,offset TxBiosVersion2
		mov	cx,33
	    rep movsb

ShowVers27:	; ask final time
		mov	ax,cs
		mov	ds,ax
 IF PCI10
		mov	ax,yx(5,13)
		mov	bx,offset TsFinalWarning
		call	PrintPosMulti
 ELSE
		mov	ax,yx(5,7)
		mov	bx,offset TsFinalWarning
		call	PrintPosMulti
		mov	ax,yx(13,7)
		mov	bx,offset TxContAbort
		call	PrintPos
 ENDIF
		call	ContinueAbort

FlashProg:	; now flash begins
	; code has little effect >>
		push	ax
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		pop	ax
	; << code has little effect
		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	al,BG_BlBlack+FG_BrWhite
		call	SetCharAttr
		mov	ax,yx(8,21)
		mov	bx,offset TxBeingUpdated
		call	PrintPos
		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(10,21)
		mov	bx,offset TsDoNotReset
		call	PrintPosMulti
		xor	bp,bp

FlashProg3:	cli
	; code has little effect >>
		mov	al,0
		call	near ptr ProgChipSet
	; << code has little effect
		mov	al,0		; BIOS shadow RAM image is swapped now with BIOS EPROM image
		call	near ptr ProgChipSet
		cmp	cs:OptionV,'V'	; erase NVRAM?
		jz	FlashProg6	; yes (image contains FFh)
		; no, copy current contents of 0EF000h-0EFFFFh
		; (plug & play info, =ACFG =ECSD) into image
		mov	ax,cs:ImageSegment
		mov	es,ax
		mov	ax,0E000h
		mov	ds,ax
		mov	si,0F000h
		mov	di,0F000h
		mov	cx,1000h
	    rep movsb

FlashProg6:	call	EnableFlashWrite
		; detect Flash EPROM device
		mov	ax,0E000h
		mov	es,ax
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Autoselect
		mov	al,es:0		; read manufacturer ID
		; 01 = Advanced Micro Devices, AMD
		cmp	al,01h		; AMD ?
		jnz	FlashProg9
		mov	al,es:1
		; 20 = Am29F010(A) 128K x 8 Sector Erase Flash Memory
		cmp	al,20h		; Am29F010(A) ?
		jnz	FlashProg9
		jmp	FlashAMD

FlashProg9:	; not Am29F010(A)
		mov	byte ptr es:5555h,0AAh	; alternate software ID entry
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Alternate
		call	Wait10m
		call	Wait10m
		call	Wait10m
		mov	al,es:0		; read manufacturer and device ID
		mov	ah,es:1
		; BF = Silicon Storage Technology, SST
		; 07 = SST29EE010 128K x 8 Page Mode EEPROM
		cmp	ax,07BFh	; SST 29EE010 ?
		jz	FlashSST
		; unknown device - just continue ?!?
		; there should be some error handling here

FlashSST:	; SST29EE010 128K x 8 Page Mode EEPROM
		; see http://www.ssti.com/products/29xe010.html
		; erase and then program chip
		mov	byte ptr es:5555h,0AAh	; software ID exit code
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Reset
		call	Wait10m
		call	Wait10m
		call	Wait10m
		mov	byte ptr es:5555h,0AAh	; software data protect disable
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_DPDisable
		call	Wait10m
		call	Wait10m
		call	Wait10m
		mov	byte ptr es:5555h,0AAh	; software chip erase
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_EraChip
		call	Wait10m
		call	Wait10m
		call	Wait10m
		cld
		mov	ax,cs:ImageSegment
		mov	ds,ax

FlashSST03:	; program 64KB per loop
		xor	si,si
		xor	di,di

FlashSST06:	; program 128 byte per loop
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Program
		mov	cx,80h
	    rep movsb				; copy 128 byte
		call	Wait300u		; T_BLCO
		dec	si
		dec	di
		lodsb				; get last written byte from image
		mov	cs:ProgTimeout,6
		; CX=0 when we come from here

FlashSST09:	; wait for completion max. 6 x 256 loops
		dec	cx
		jnz	FlashSST12
		dec	cs:ProgTimeout
		jnz	FlashSST12
		jmp	EBadFlash		; problem

FlashSST12:	mov	ah,es:[di]		; get status byte from flash chip
		cmp	al,ah			; same == done?
		jnz	FlashSST09		; no - check again
		inc	di			; to next 128 byte
		or	di,di			; 64 KB done?
		jnz	FlashSST06		; no - continue
		; 64KB done
		mov	ax,es
		cmp	ax,0F000h		; did we just do the second 64 KB?
		jz	FlashSST15		; yes - done
		add	ax,1000h		; no, advance segment
		mov	es,ax
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
		jmp	short FlashSST03	; next (final) 64 KB

FlashSST15:	; done
		call	Wait10m
		jmp	short FlashComp

FlashAMD:	; Am29F010(A) 128K x 8 Sector Erase Flash Memory
		; see http://www.amd.com/products/nvd/techdocs/techdocs.html#29Fseries
		; ensure that all sectors are writeable, then erase and program chip
		mov	cx,8
		mov	bx,2

FlashAMD03:	or	bx,2
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Autoselect
		mov	al,es:[bx]
		cmp	al,1		; sector protected?
		jnz	FlashAMD06
		jmp	EWriteProtect	; write protect error

FlashAMD06:	cmp	al,0
		jz	FlashAMD09
		jmp	EBadFlash	; problem

FlashAMD09:	and	bx,0F000h
		add	bx,4000h
		or	bx,bx
		jnz	FlashAMD12
		mov	ax,0F000h
		mov	es,ax
FlashAMD12:	loop	FlashAMD03

FlashAMD15:	mov	ax,cs:ImageSegment
		mov	ds,ax
		call	FProgAMD
		jb	FlashAMD15
		xor	cx,cx
		loop	$		; wait a bit

FlashComp:	; programming Am29F010 or SST29EE010 complete
		; compare memory image against programmed chip
		mov	bx,0

FlashComp3:	mov	ds,bx
		xor	si,si
		xor	di,di
		mov	cx,8000h
	    rep lodsw
		add	bh,10h
		cmp	bh,0A0h
		jnz	FlashComp3
		; compare Flash EPROM
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,0E000h
		mov	es,ax
		xor	di,di
		xor	si,si
		mov	cx,8000h
	    rep cmpsw
		jnz	FlashComp6
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
		mov	ax,0F000h
		mov	es,ax
		xor	di,di
		xor	si,si
		mov	cx,8000h
	    rep cmpsw
		jnz	FlashComp6

		; flash program went ok, now halt machine (wait for power off)
		mov	al,1
		call	near ptr ProgChipSet
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	bx,offset TsSuccess
		mov	ax,yx(6,2)
		call	PrintPosMulti
		cli
		loop	$
		loop	$
		loop	$
		hlt

FlashComp6:	; compare failed
		inc	cs:ProgRetries
		cmp	cs:ProgRetries,5
		jz	EFatalError
		jmp	FlashProg3	; retry

EFatalError:	; fatal error (compare failed 5 times)
		mov	al,1
		call	near ptr ProgChipSet
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	al,BG_Red+FG_White
		call	SetCharAttr
		mov	ax,yx(3,6)	; from
		mov	cx,yx(17,70)	; to
		call	ClearScreenArea
		mov	ax,yx(6,11)
		mov	bx,offset TsFatalError
		call	PrintPosMulti
		call	CacheOn
		jmp	TerminateProg

EBadFlash:	; programming problem
		call	DisableFlashWrite
		mov	al,1
		call	near ptr ProgChipSet
		call	CacheOn
		sti
		mov	bx,offset TxBadFlashRom
		jmp	PrintTermProg

EWriteProtect:	; write protect
		call	CacheOn
		call	DisableFlashWrite
		mov	al,1
		call	near ptr ProgChipSet
		mov	bx,offset TxWriteProtect
		jmp	PrintTermProg

;============================================================================

;		S u b r o u t i n e

CacheOn		proc near
		mov	si,0B001h
		mov	al,1
		nxop
		retn
CacheOn		endp


;============================================================================

;		S u b r o u t i n e
; Wait 300 usec (30 x 10 usec)

Wait300u	proc near
		push	cx
		mov	cx,30
Wait300u3:	call	Wait10u
		loop	Wait300u3
		pop	cx
		retn
Wait300u	endp


;============================================================================

;		S u b r o u t i n e
; Wait 6 usec

Wait6u		proc near
		push	cx
		mov	cx,cs:DelayFact6
		loop	$
		pop	cx
		retn
Wait6u		endp


;============================================================================

;		S u b r o u t i n e
; Wait 10 usec

Wait10u		proc near
		push	cx
		mov	cx,cs:DelayFact10
		loop	$
		pop	cx
		retn
Wait10u		endp


;============================================================================

;		S u b r o u t i n e
; Wait 10 msec (1000 x 10 usec)

Wait10m		proc near
		push	cx
		push	ax
		mov	ax,1000
Wait10m3:	mov	cx,cs:DelayFact10
		loop	$
		dec	ax
		jnz	Wait10m3
		pop	ax
		pop	cx
		retn
Wait10m		endp


; ===========================================================================

;		S u b r	o u t i	n e

 IF PCI10
GetBiosFName	proc near
		push	eax
		push	dx
		mc3	<mov di,OBiosFilename>
		pop	dx
		pop	eax
		retn
GetBiosFName	endp
 ENDIF

;============================================================================

;		S u b r o u t i n e
; clear and program AMD Am29F010

FProgAMD	proc near
		mov	ax,0E000h
		mov	es,ax
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_EraChip

FProgAMD03:	mov	al,es:0
		test	al,80h
		jnz	FProgAMD06
		test	al,20h
		jnz	FProgAMD
		jmp	short FProgAMD03

FProgAMD06:	mov	ax,0E000h
		mov	es,ax
		cld

FProgAMD09:	mov	cx,0
		xor	bx,bx
		xor	si,si

FProgAMD12:	lodsb
		mov	ah,al
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Program
		mov	es:[bx],ah
		mov	dh,ah
		and	dh,80h

FProgAMD15:	mov	al,es:[bx]
		and	al,80h
		cmp	al,dh
		jnz	FProgAMD15
		mov	al,es:[bx]
		cmp	al,ah
		jnz	FProgAMD15
		inc	bx
		loop	FProgAMD12
		mov	ax,es
		cmp	ax,0F000h
		jz	FProgAMD18
		add	ax,1000h
		mov	es,ax
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
		jmp	short FProgAMD09

FProgAMD18:	dec	bx
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Reset
		mov	al,es:[bx]
		clc
		retn
FProgAMD	endp


;============================================================================

;		S u b r o u t i n e

Read32KFromFile proc near
		mov	cx,8000h
		push	ds
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ah,3Fh
		int	21h		; DOS - 2+ - READ FROM FILE WITH HANDLE
					; BX = file handle, CX = number of bytes to read
					; DS:DX -> buffer
		pop	ds
		retn
Read32KFromFile endp

	; unused code >>
		jnb	Read32Trash
		pop	ax
		jmp	short EFileSize

Read32Trash:	cmp	ax,8000h
		jz	Read32Trash3
		pop	ax
		jmp	short EFileSize

Read32Trash3:	retn
	; << unused code

;============================================================================

;		S u b r o u t i n e
; Add all words of 128K BIOS image
; E: ImageSegment
; R: Carry if check failed, NCarry if check ok

CheckChecksum	proc near
		push	ds
		cld
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	cx,8000h
		xor	bx,bx
		xor	ax,ax
ChkChksum3:	add	ax,[bx]
		add	bx,2
		loop	ChkChksum3
		mov	bx,ds
		add	bx,1000h
		mov	ds,bx
		mov	cx,8000h
		xor	bx,bx
ChkChksum6:	add	ax,[bx]
		add	bx,2
		loop	ChkChksum6
		pop	ds
		cmp	ax,0		; ok?
		jnz	ChkChksum9	; no
		; check ok
		clc
		retn

ChkChksum9:	; check failed
		stc
		retn
CheckChecksum	endp


; File problems handled here: close file, tell user,
; terminate program or ask for different filename

		assume ds:flashp

	; unused code >>
		mov	bx,FileHandle1
		call	CloseFile
		mov	bx,offset TxWrongSystem
		jmp	short EPrintTerm
	; << unused code

ENoNexGenBios:	mov	bx,FileHandle1
		call	CloseFile
		mov	bx,offset TxNoNexGenBios
		jmp	short EPrintTerm

EFileSize:	mov	bx,FileHandle1
		call	CloseFile
		mov	bx,offset TxFileSizeErr
		jmp	short EPrintTerm

EFileNotExist:	mov	bx,offset TxFileNotExist
EPrintTerm:	; print and terminate program
		mov	al,BG_Blue+FG_Red
		call	SetCharAttr
		mov	ax,yx(8,7)
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		call	GetKeyb
 IF not PCI10
		jmp	TerminateProg
 ENDIF
	; unused code if not PCI10 >>
		cmp	cs:OptionF,'F'
		jz	ETerm
		jmp	Main18		; ask again for filename
	; << unused code if not PCI10

ETerm:		jmp	PrintTermProg	; why print again??

EClosePrTerm:	; close file, print error message and terminate program
		push	bx
		mov	bx,FileHandle1
		call	CloseFile
		pop	bx
		jmp	short EPrintTerm

	; unused code >>
AskDest:	; ask for new destination filename
		mov	ax,cs
		mov	ds,ax
		mov	ax,yx(13,1)	; from
		mov	cx,yx(18,78)	; to
		call	ClearScreenArea
		mov	al,BG_Blue+FG_Red
		call	SetCharAttr
		mov	ax,yx(14,6)	; from
		mov	cx,yx(16,70)	; to
		call	ClearScreenArea
		mov	ax,yx(13,7)
		mov	bx,offset TxEnterSave
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		; cut trailing space chars from TxSaveFilename
		mov	bx,TxSaveFilename
AskDest2:	cmp	byte ptr [bx],0
		jz	AskDest4
		inc	bx
		jmp	short AskDest2

AskDest4:	dec	bx
		cmp	byte ptr [bx],' '
		jz	AskDest4
		inc	bx
		mov	byte ptr [bx],0
		; get new filename from user
		push	bx
		mov	bx,TxSaveFilename
		mov	ax,yx(15,7)
		call	PrintPos
		pop	bx
		mov	si,offset TxSaveFilename-1
		mov	di,offset TxSaveFilename-1+EditBufSize
		call	Editor
		; try to make backup
		mov	dx,offset TxSaveFilename
		call	CreateFile
		jb	ECantCreate
		mov	FileHandle2,ax
		mov	bx,ax
		mov	ax,es:ImageSegment
		mov	ds,ax
		mov	cx,0FFFFh
		mov	dx,0
		call	WriteFile
		jb	EFileWrite
		call	FilePosFFFF
		mov	dx,0FFFFh
		mov	cx,1
		mov	bx,es:FileHandle2
		call	WriteFile
		jb	EFileWrite
		mov	bx,offset TxUpdComplete
	; << unused code

PrintTermProg:	; output message and terminate program
		push	bx
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		pop	bx
		mov	ax,yx(6,16)
		call	PrintPos
		mov	dx,yx(24,0)
		call	SetCursor
		jmp	short TerminateProg

	; unused code >>
ECantCreate:	mov	bx,offset TxCantCreate
		jmp	short EBackup

EFileWrite:	mov	bx,FileHandle2
		call	CloseFile
		mov	bx,offset TxFileWriteErr

EBackup:	; problem with backup file
		; ask user for a new destination filename
		mov	al,BG_Blue+FG_Magenta
		call	SetCharAttr
		mov	ax,yx(17,7)
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		call	GetKeyb
		jmp	AskDest
	; << unused code

TerminateProg:	sti
		mov	ax,4C01h	; QUIT WITH EXIT CODE (AL = code)
		int	21h		; DOS
		jmp	short $		; halt system (never reached)

;============================================================================

;		S u b r o u t i n e

OpenFile	proc near
		mov	ah,3Dh
		mov	al,0
		int	21h		; DOS - 2+ - OPEN DISK FILE WITH HANDLE
					; DS:DX -> ASCIZ filename
					; AL = access mode
					; 0 - read
		retn
OpenFile	endp


;============================================================================

;		S u b r o u t i n e

CloseFile	proc near
		mov	ah,3Eh
		int	21h		; DOS - 2+ - CLOSE A FILE WITH HANDLE
					; BX = file handle
		retn
CloseFile	endp


;============================================================================

;		S u b r o u t i n e

CreateFile	proc near
		mov	ah,3Ch
		xor	cx,cx
		int	21h		; DOS - 2+ - CREATE A FILE WITH HANDLE (CREAT)
					; CX = attributes for file
					; DS:DX -> ASCIZ filename (may include drive and path)
		retn
CreateFile	endp


;============================================================================

;		S u b r o u t i n e

WriteFile	proc near
		mov	ah,40h
		int	21h		; DOS - 2+ - WRITE TO FILE WITH HANDLE
					; BX = file handle, CX = number of bytes to write, DS:DX -> buffer
		retn
WriteFile	endp


;============================================================================

;		S u b r o u t i n e

FilePosFFFF	proc near
		mov	ax,4200h
		mov	cx,0
		mov	dx,0FFFFh
		int	21h		; DOS - 2+ - MOVE FILE READ/WRITE POINTER (LSEEK)
					; AL = method: offset from beginning of file
		retn
FilePosFFFF	endp


;============================================================================

;		S u b r o u t i n e
; E: pointer into buffer in BX, start of buffer in SI, end of buffer in DI
; R: buffer (no return if ESC key pressed)

	; unused code >>
Editor03:	inc	bx		; undo dec

Editor		proc near
		call	GetKeyb
		cmp	al,CR		; done?
		jnz	Editor06	; no, edit
		retn
Editor		endp
	; << unused code


;============================================================================

;		S u b r o u t i n e
; E:-
; R:- (No return if ESC key pressed)

ContAbor6:	; handle abort
		pop	ax		; drop return address
		mov	bx,offset TxAbortUpdate
		jmp	short PrintTermProg

ContinueAbort	proc near
		call	GetKeyb
		and	al,0DFh		; convert to upper case
		cmp	al,'C'		; continue?
		jnz	ContAbor3
		retn

ContAbor3:	cmp	al,ESCP		; abort?
		jnz	ContinueAbort
		jmp	short ContAbor6
ContinueAbort	endp


;============================================================================

; extension for subroutine Editor
; AL= ascii char, AH= scan code
; BX= pointer into buffer
; SI= start of buffer, DI= end of buffer

	; unused code >>
Editor06:	cmp	al,ESCP		; abort?
		jnz	Editor09	; no
		; handle abort
		mov	ax,cs
		mov	ds,ax
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		pop	ax		; drop return address
		jmp	short TerminateProg

Editor09:	cmp	ah,53h		; delete key?
		jz	Editor30
		cmp	ah,52h		; insert key?
		jnz	Editor12
		jmp	Editor48

Editor12:	cmp	al,BS		; backspace
		jz	Editor18
		cmp	ah,4Bh		; cursor left key?
		jz	Editor18
		cmp	ah,4Dh		; cursor right key?
		jz	Editor21
		cmp	al,' '		; control char?
		jb	Editor	; ignore
		cmp	al,'@'		; < A-Za-z?
		jb	Editor15
		cmp	al,DEL		; > DEL?
		ja	Editor		; ignore

Editor15:	; put char into buffer and echo it
		cmp	bx,di		; end of buffer reached?
		jz	Editor		; yes, ignore
		mov	[bx],al		; no, put in
		mov	es:TxTmpChar,al
		push	si
		mov	si,offset TxTmpChar ; print char
		call	PrintBS
		pop	si
		inc	bx
		jmp	short Editor

Editor18:	; handle backspace / cursor left
		dec	bx
		cmp	bx,si		; start of buffer reached?
		jz	Editor03
		push	si
		mov	si,offset TxBackSpace
		call	PrintBS		; print BS
		pop	si
		jmp	short Editor

Editor21:	; handle cursor right
		mov	al,[bx]
		cmp	al,0		; if at end of string
		jnz	Editor24
		mov	al,' '		; enlarge string with space
Editor24:	mov	es:TxTmpChar,al
		inc	bx
		cmp	bx,di		; end of buffer reached?
		jz	Editor27
		push	si
		mov	si,offset TxTmpChar
		call	PrintBS		; print char
		pop	si
		jmp	Editor

Editor27:	dec	bx		; undo inc
		jmp	Editor

Editor30:	; handle delete
		cmp	bx,di		; at end of buffer?
		jnz	Editor33
		jmp	short Editor45

Editor33:	push	bx
Editor36:	inc	bx
		cmp	bx,di
		jz	Editor39
		mov	al,[bx]
		dec	bx
		mov	[bx],al
		inc	bx
		cmp	bx,di
		jz	Editor39
		cmp	al,0
		jnz	Editor36
		dec	bx
		mov	byte ptr [bx],' '
		inc	bx
		mov	byte ptr [bx],' '
		jmp	short Editor42

Editor39:	dec	bx
		mov	byte ptr [bx],' '
Editor42:	pop	bx
		push	bx
		push	si
		mov	si,bx
		call	GetCursor
		push	dx
		call	Print
		pop	dx
		call	SetCursor	; E: xy in DX
		pop	si
		pop	bx
Editor45:	jmp	Editor

Editor48:	; handle insert
		cmp	bx,di		; at end of buffer?
		jz	Editor45
		push	bx
		mov	ah,' '
		mov	al,[bx]
		mov	[bx],ah
Editor51:	inc	bx
		cmp	bx,di
		jz	Editor54
		mov	ah,[bx]
		mov	[bx],al
		cmp	al,0
		jz	Editor54
		inc	bx
		cmp	bx,di
		jz	Editor54
		mov	al,[bx]
		mov	[bx],ah
		cmp	ah,0
		jz	Editor54
		jmp	short Editor51

Editor54:	pop	bx
		push	bx
		push	si
		mov	si,bx
		call	GetCursor
		push	dx
		call	Print
		pop	dx
		call	SetCursor	; E: xy in DX
		pop	si
		pop	bx
		jmp	Editor

TxTmpChar	db 0,0			; temporary storage to print one char
TxBackSpace	db BS,0
	; << unused code

;============================================================================

;		S u b r o u t i n e
; E: xy in AX

SetCursorSav	proc near
		push	ax
		push	bx
		mov	dx,ax
		call	SetCursor	; E: xy in DX
		pop	bx
		pop	ax
		retn
SetCursorSav	endp


;============================================================================

;		S u b r o u t i n e
; E: xy in DX

SetCursor	proc near
		mov	ah,2
		mov	bh,0
		int	10h		; - VIDEO - SET CURSOR POSITION
					; DH,DL = row, column (0,0 = upper left)
					; BH = page number
		retn
SetCursor	endp


;============================================================================

;		S u b r o u t i n e

GetCursor	proc near
		mov	ah,3
		mov	bh,0
		int	10h		; - VIDEO - READ CURSOR POSITION
					; BH = page number
					; Return: DH,DL = row,column, CH = cursor start line, CL = cursor end line
		retn
GetCursor	endp


;============================================================================

;		S u b r o u t i n e
; Another print routine. Sigh!
; Prints at the current cursor position. This time it handles backspaces too.
; E: pointer to string in SI
; R: SI points after terminating 0, AH=14

PrintBS		proc near
		push	di
		push	bx
		xor	bh,bh
PrintBS3:	mov	ah,14		; why??
		lods	byte ptr es:[si]
		test	al,al		; =0?
		jz	PrintBS9
		cmp	al,BS		; backspace?
		jnz	PrintBS6	; no
		; handle backspace
		call	GetCursor
		dec	dx		; back one X-position
		call	SetCursor
		jmp	short PrintBS3

PrintBS6:	; print char in AL
		push	ax
		call	GetCursor
		pop	ax
		push	dx
		push	bx
		mov	bh,0		; display page
		mov	bl,BG_Blue+FG_White
		mov	cx,1		; number of times to write char
		mov	ah,9		; WRITE ATTRIBUTES/CHARACTERS AT CURSOR POSITION
		int	10h		; VIDEO
		pop	bx
		pop	dx
		inc	dx		; forward one X-position
		call	SetCursor
		jmp	short PrintBS3

PrintBS9:	pop	bx
		pop	di
		retn
PrintBS		endp


;============================================================================

;		S u b r o u t i n e
; E: pointer to string in SI
; R: SI points after terminating 0
Print		proc near
		push	di
		push	bx
		xor	bh,bh		; display page 0

Print3:		mov	ah,14		; WRITE CHARACTER AND ADVANCE CURSOR (TTY WRITE)
		lodsb
		test	al,al		; terminating 0?
		jz	Print6		; yes, done
		int	10h		; VIDEO
					; AL = character, BH = display page (alpha modes)
					; BL = foreground color (graphics modes)
		jmp	short Print3

Print6:		pop	bx
		pop	di
		retn
Print		endp


;============================================================================

;		S u b r o u t i n e
; E: xy in AX, pointer to string in BX
; R: BX points to terminating 0

PrintPos	proc near
		push	ax
		push	bx
		push	dx
		push	cx
		push	di
		call	SetCursorSav	; E: xy in AX
PrintPo2:	push	ax
		mov	al,[bx]
		cmp	al,0
		jz	PrintPo4
		push	bx
		mov	bh,0
		mov	bl,es:CharAttr
		mov	cx,1
		mov	ah,9		; WRITE ATTRIBUTES/CHARACTERS AT CURSOR POSITION
		int	10h		; VIDEO
					; AL = character, BH = display page
					; BL = attributes of character (alpha modes) or color (graphics modes)
					; CX = number of times to write character
		pop	bx
		pop	ax
		inc	ax
		call	SetCursorSav	; E: xy in AX
		inc	bx
		jmp	short PrintPo2

PrintPo4:	pop	ax
		pop	di
		pop	cx
		pop	dx
		pop	bx
		pop	ax
		retn
PrintPos	endp


;============================================================================

;		S u b r o u t i n e
; Print a list of strings, each one at a new line
; E: xy in AX, pointer to string in BX

PrintPosMulti	proc near
		call	PrintPos
		inc	ah		; increment Y-position
PrintPosM2:	inc	bx		; skip terminating 0
		cmp	byte ptr cs:[bx],0
		jnz	PrintPosM2
		inc	bx		; goto next string
		cmp	byte ptr cs:[bx],0 ; done?
		jnz	PrintPosMulti	; no, print it
		retn
PrintPosMulti	endp

	; unused code >>
		call	PrintRectangle  ; E: xy upper left corner in AX, xy lower right corner in CX
		retn
	; << unused code

;============================================================================

;		S u b r o u t i n e
; E: xy upper left corner in AX, xy lower right corner in CX

PrintRectangle	proc near
		push	ax
		push	cx
		inc	al
		sub	cl,al
		xor	ch,ch
		push	cx
		push	ax
		call	PrintLinePos	; E: xy in AX, number of spaces in CX
		pop	ax
		pop	cx
		pop	dx
		push	dx
		mov	ah,dh
		call	PrintLinePos	; E: xy in AX, number of spaces in CX
		pop	cx
		pop	ax
		push	ax
		push	cx
		mov	bx,offset TxUppLeftCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
PrintRect2:	mov	bx,offset TxVertLine
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
		cmp	ah,ch
		jnz	PrintRect2
		mov	bx,offset TxLwrLeftCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		pop	cx
		pop	ax
		mov	al,cl
		mov	bx,offset TxUppRghtCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
PrintRect4:	mov	bx,offset TxVertLine
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
		cmp	ah,ch
		jnz	PrintRect4
		mov	bx,offset TxLwrRghtCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		retn
PrintRectangle	endp


;============================================================================

;		S u b r o u t i n e
; Print a line at xy position AX, length CX
; E: xy in AX, number of line-chars in CX

PrintLinePos	proc near
		mov	bx,offset TxHorizLine
		call	PrintPos
		inc	al		; increment X-position
		loop	PrintLinePos
		retn
PrintLinePos	endp


;============================================================================

;		S u b r o u t i n e
; Prints 3 strings on same line at X-positions AL, 39 and 72
; E: xy in AX, pointer to 3 strings in BX

	; unused code >>
Print3Str	proc near
		call	PrintPos
		mov	al,39		; X-position for second string
		call	PrintPos
		mov	al,72		; X-position for third string
		call	PrintPos
		retn
Print3Str	endp
	; << unused code


;============================================================================

;		S u b r o u t i n e
; E: xy upper left corner in AX, xy lower right corner in CX
; R:-

ClearScreenArea proc near
		push	ax
		push	cx
		inc	al
		sub	cl,al		; width of rectangle to clear
		xor	ch,ch
		mov	bx,offset TxSpaceLine+82
		sub	bx,cx		; BX is pointer to space-char
		pop	cx		; string of required length
		pop	ax
		push	ax
		inc	ah
		sub	ch,ah		; heigh of rectangle to clear
		xchg	ch,cl
		xor	ch,ch
		add	cx,2		; CX is number of lines to clear
		pop	ax		; position of upper left corner
ClrScrArea2:	push	bx
		call	PrintPos
		add	ax,yx(1,0)	; increment Y-coord
		pop	bx
		loop	ClrScrArea2
		retn
ClearScreenArea endp


;============================================================================

;		S u b r o u t i n e
; E:-
; R: AH = scan code, AL = character

GetKeyb		proc near
		push	si
		push	di
		push	bx
		mov	ah,0		; READ CHAR FROM BUFFER, WAIT IF EMPTY
		int	16h		; KEYBOARD
					; Return: AH = scan code, AL = character
		pop	bx
		pop	di
		pop	si
		retn
GetKeyb		endp


;============================================================================

;		S u b r o u t i n e
; E: character attribute in AL (ES must be CS)
; V:- / R:-

SetCharAttr	proc near
		mov	es:CharAttr,al
		retn
SetCharAttr	endp


;============================================================================

;		S u b r o u t i n e
; Enable flash EPROM programming
; E:- / V: AL / R:-

EnableFlashWrite proc near		; set keyboard controller P12 line low
		mov	al,0B2h
		out	64h,al		; AT Keyboard controller
		retn
EnableFlashWrite endp


	; unused code >>
KbdTrash:	in	al,64h		; AT Keyboard controller
		test	al,2		; input buffer full?
		jnz	KbdTrash
		retn
		retn			; even?
	; << unused code

;============================================================================

;		S u b r o u t i n e
; Disable flash EPROM programming
; E:- / V: AX, CX, Flags / R:-

DisableFlashWrite proc near		; set keyboard controller P12 line high
		mov	al,0BAh
		call	KbdOutComByte	; E: keyboard command byte in AL
					; V: AX, CX, Flags
					; R: NZero if timeout
		call	KbdOut64Inp60	; E: keyboard controller command in AL, expected value in AH
					; V: AL, CX, Flags
					; R: value in AL, NZero if timeout or value not AH
		retn
DisableFlashWrite endp


;============================================================================

;		S u b r o u t i n e

ShowFileList	proc near
		mov	dx,cs
		mov	ds,dx
		mov	dx,offset FileNameBuffer
		mov	ah,1Ah
		int	21h		; DOS - SET DISK TRANSFER AREA ADDRESS
					; DS:DX -> disk transfer buffer
		mov	ax,yx(6,2)
		push	ax
		mov	cx,0
		mov	dx,cs
		mov	ds,dx
		mov	dx,offset TxDotStarDot
		mov	ah,4Eh
		int	21h		; DOS - 2+ - FIND FIRST ASCIZ (FINDFIRST)
					; CX = search attributes
					; DS:DX -> ASCIZ filespec
					; (drive, path, and wildcards allowed)
		jb	ShowFileL6

ShowFileL2:	mov	bx,offset FileNameBuffer+1Eh
		pop	ax
		push	ax
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		pop	ax
		add	ax,yx(1,0)
		push	ax
		cmp	ah,13
		jnz	ShowFileL4
		pop	ax
		mov	ah,6
		add	al,14
		push	ax
		cmp	al,66
		jbe	ShowFileL4
		jmp	short ShowFileL6

ShowFileL4:	mov	ah,4Fh
		int	21h		; DOS - 2+ - FIND NEXT ASCIZ (FINDNEXT)
					; [DTA] = data block from
					; last AH = 4Eh/4Fh call
		jb	ShowFileL6
		jmp	short ShowFileL2

ShowFileL6:	pop	ax
		retn
ShowFileList	endp

TxDotStarDot	db '*.*',0
TxBiosDotOld	db 'bios.old',0

;============================================================================

;		S u b r o u t i n e

SaveOldBios	proc near
		mov	ax,cs
		mov	ds,ax
		mov	dx,offset TxBiosDotOld
		mov	ah,3Dh
		mov	al,2
		int	21h		; DOS - 2+ - OPEN DISK FILE WITH HANDLE
					; DS:DX -> ASCIZ filename
					; AL = access mode
					; 2 - read & write
		jb	SaveOldB2
		mov	dx,offset TxBiosDotOld
		mov	ah,41h
		int	21h		; DOS - 2+ - DELETE A FILE (UNLINK)
					; DS:DX -> ASCIZ pathname of file to delete (no wildcards allowed)

SaveOldB2:	mov	dx,offset TxBiosDotOld
		call	CreateFile
		jb	SaveOldB4
		mov	cs:FileHandle1,ax
		mov	bx,ax
		xor	dx,dx
		mov	ax,cs:ImageSegment
		call	Write32KToFile
		jb	SaveOldB6
		mov	dx,8000h
		mov	ax,cs:ImageSegment
		call	Write32KToFile
		jb	SaveOldB6
		xor	dx,dx
		mov	ax,cs:ImageSegment
		add	ax,1000h
		call	Write32KToFile
		jb	SaveOldB6
		mov	dx,8000h
		mov	ax,cs:ImageSegment
		add	ax,1000h
		call	Write32KToFile
		jb	SaveOldB6
		clc
		retn
SaveOldBios	endp


;============================================================================

;		S u b r o u t i n e

Write32KToFile	proc near
		push	ds
		push	ax
		mov	cx,8000h
		mov	ds,ax
		pop	ax
		mov	ah,40h
		int	21h		; DOS - 2+ - WRITE TO FILE WITH HANDLE
					; BX = file handle, CX = number of bytes to write, DS:DX -> buffer
		pop	ds
		retn
Write32KToFile	endp

SaveOldB4:	mov	si,offset TxCreateErr
		jmp	short SaveOldB8

SaveOldB6:	mov	si,offset TxWriteFileErr
SaveOldB8:	call	Print
		stc
		retn

;============================================================================

;		S u b r o u t i n e
; E: xy in AX

PrintDot	proc near
		push	bx
		mov	bx,offset TxDot
		call	PrintPos
		pop	bx
		retn
PrintDot	endp

;============================================================================

;		S u b r o u t i n e
; Swap BIOS shadow RAM image with BIOS EPROM image
; E: 0 in AL to map EPROM image, 1 in AL to map shadow RAM image

ProgChipSet	proc near
		pushad
		cmp	al,0		; map EPROM image?
		jnz	ProgChipS2	; no

		; map EPROM image
		mov	dx,0CF8h	; reset NxPCI 0x58.0
		mov	eax,80000058h
		out	dx,eax
		mov	dl,0FCh
		in	eax,dx
		and	eax,0FFFFFFFEh
		out	dx,eax

		mov	dx,0FCF8h	; reset NxMC 0x08 uppermost byte
		mov	eax,80000008h
		out	dx,eax
		mov	dl,0FCh
		in	eax,dx
		and	eax,0FFFFFFh
		out	dx,eax

		mov	dx,0FCF8h	; reset NxMC 0x0C uppermost byte
		mov	eax,8000000Ch
		out	dx,eax
		mov	dl,0FCh
		in	eax,dx
		and	eax,0FFFFFFh
		out	dx,eax

		mov	dx,0CF8h	; reset NxPCI 0x4C.25 - .16
		mov	eax,8000004Ch
		out	dx,eax
		mov	dl,0FCh
		in	eax,dx
		and	eax,0FC00FFFFh
		out	dx,eax
		jmp	short ProgChipS4

ProgChipS2:	; map shadow RAM image
		mov	dx,0FCF8h	; set NxMC 0x08 uppermost byte
		mov	eax,80000008h
		out	dx,eax
		mov	dl,0FCh
		in	eax,dx
		or	eax,0FF000000h
		out	dx,eax
		mov	dx,0FCF8h	; no effect

ProgChipS4:	; done
		popad
		retn
ProgChipSet	endp


;============================================================================

;		S u b r o u t i n e
; E: keyboard controller command in AL, expected value in AH
; V: AL, CX, Flags
; R: value in AL, NZero if timeout or value not AH

KbdOut64Inp60	proc near		; E: keyboard controller command in AL
		call	KbdOut64	; V: CX, Flags
					; R: NZero if timeout
		jnz	KbdO64I602
		call	KbdInp60	; V: AL, CX, Flags
					; R: value in AL, Zero if timeout
		jz	KbdOutI604
		cmp	al,ah
KbdO64I602:	retn
KbdOut64Inp60	endp


;============================================================================

;		S u b r o u t i n e
; E: keyboard command in AL, expected value in AH
; V: AL, CX, Flags
; R: value in AL, NZero if timeout or value not AH

KbdOut60Inp60	proc near		; E: keyboard command in AL
		call	KbdOut60	; V: AL, CX, Flags
					; R: NZero if timeout
		jnz	KbdO60I602
		call	KbdInp60	; V: AL, CX, Flags
					; R: value in AL, Zero if timeout
		jz	KbdOutI604
		cmp	al,ah
KbdO60I602:	retn

KbdOutI604:	or	sp,sp		; set NZero
		retn
KbdOut60Inp60	endp


;============================================================================

;		S u b r o u t i n e

ReadKbdInpPort	proc near		; AT Keyboard controller 8042.
		in	al,64h
		test	al,1
		jz	ReadKbdI4
		jcxz	ReadKbdI2
ReadKbdI2:	in	al,60h		; AT Keyboard controller 8042.
ReadKbdI4:	mov	al,0C0h		; read input port and place on port 60h
		call	KbdOut64Inp60	; E: keyboard controller command in AL, expected value in AH
					; V: AL, CX, Flags
					; R: value in AL, NZero if timeout or value not AH
		retn
ReadKbdInpPort  endp


;============================================================================

;		S u b r o u t i n e

ReadKbdOutPort	proc near		; AT Keyboard controller 8042.
		in	al,64h
		test	al,1
		jz	ReadKbdO4
		jcxz	ReadKbdO2

ReadKbdO2:	in	al,60h		; AT Keyboard controller 8042.
ReadKbdO4:	mov	al,0D0h		; read output port and place on port 60h
		call	KbdOut64Inp60	; E: keyboard controller command in AL, expected value in AH
					; V: AL, CX, Flags
					; R: value in AL, NZero if timeout or value not AH
		retn
ReadKbdOutPort  endp


;============================================================================

;		S u b r o u t i n e
; V: AL, CX, Flags
; R: value in AL, Zero if timeout

KbdInp60	proc near		; V: AL, CX, Flags
		call	KbdWaitOutBuf	; R: Zero if timeout
		in	al,60h		; AT Keyboard controller 8042.
		retn
KbdInp60	endp


;============================================================================

;		S u b r o u t i n e
; V: AL, CX, Flags
; R: Zero if timeout

KbdWaitOutBuf	proc near
		mov	cx,12
KbdWaitO2:	push	cx
		xor	cx,cx
KbdWaitO4:	in	al,64h		; AT Keyboard controller 8042.
		test	al,1		; output buffer full?
		loope	KbdWaitO4
		pop	cx
		loope	KbdWaitO2
		retn
KbdWaitOutBuf	endp


;============================================================================

;		S u b r o u t i n e
; V: AX, CX, Flags
; R: value in AL, Zero if timeout

KbdInp60A	proc near
		mov	ah,12
		xor	cx,cx
KbdInp60A2:	in	al,64h		; AT Keyboard controller 8042.
		test	al,1
		loope	KbdInp60A2
		jnz	KbdInp60A4
		dec	ah
		jnz	KbdInp60A2
KbdInp60A4:	in	al,60h		; AT Keyboard controller 8042.
		retn
KbdInp60A	endp


;============================================================================

;		S u b r o u t i n e
; E: keyboard controller command in AL, second byte in AH
; V: AL, CX, Flags
; R: NZero if timeout

KbdOut6460	proc near		; E: keyboard controller command in AL
		call	KbdOut64	; V: CX, Flags
					; R: NZero if timeout
		jnz	KbdOut64602
		mov	al,ah
		call	KbdOut60	; E: keyboard command in AL
					; V: AL, CX, Flags
					; R: NZero if timeout
KbdOut64602:	retn
KbdOut6460	endp


;============================================================================

;		S u b r o u t i n e
; E: port value in AL
; V: AX, CX, Flags
; R: NZero if timeout

WriteKbdOutPort	proc near
		mov	ah,al
		mov	al,0D1h		; write output port.
					; The next byte written to port 60h
					; will be written to the 804x output port
		call	KbdOut6460	; E: keyboard controller command in AL, second byte in AH
					; V: AL, CX, Flags
					; R: NZero if timeout
		retn
WriteKbdOutPort	endp


;============================================================================

;		S u b r o u t i n e
; E: keyboard controller command in AL
; V: CX, Flags
; R: NZero if timeout

KbdOut64	proc near
		push	ax
		call	KbdWaitInpBuf	; V: AL, CX, Flags
					; R: NZero if timeout
		pop	ax
		jnz	KbdOut642
		out	64h,al		; AT Keyboard controller 8042.
		push	ax
		call	KbdWaitInpBuf	; V: AL, CX, Flags
					; R: NZero if timeout
		pop	ax
KbdOut642:	retn
KbdOut64	endp


;============================================================================

;		S u b r o u t i n e
; E: keyboard command in AL
; V: AL, CX, Flags
; R: NZero if timeout

KbdOut60	proc near
		push	ax
		call	KbdWaitInpBuf	; V: AL, CX, Flags
					; R: NZero if timeout
		pop	ax
		jnz	KbdOut602
		out	60h,al		; AT Keyboard controller 8042.
		call	KbdWaitInpBuf	; V: AL, CX, Flags
					; R: NZero if timeout
KbdOut602:	retn
KbdOut60	endp


;============================================================================

;		S u b r o u t i n e
; E: keyboard command byte in AL
; V: AX, CX, Flags
; R: NZero if timeout

KbdOutComByte	proc near
		mov	ah,al
		mov	al,60h		; load new command
		call	KbdOut6460	; E: keyboard controller command in AL, second byte in AH
					; V: AL, CX, Flags
					; R: NZero if timeout
		retn
KbdOutComByte	endp


;============================================================================

;		S u b r o u t i n e
; V: AL, CX, Flags
; R: NZero if timeout

KbdWaitInpBuf	proc near
		xor	cx,cx
KbdWaitI2:	in	al,64h		; AT Keyboard controller 8042.
		test	al,2		; input buffer full?
		loopne	KbdWaitI2
		jz	KbdWaitI8
KbdWaitI4:	in	al,64h		; AT Keyboard controller 8042.
		test	al,2		; input buffer full?
		loopne	KbdWaitI4
		jz	KbdWaitI8
KbdWaitI6:	in	al,64h		; AT Keyboard controller 8042.
		test	al,2		; input buffer full?
		loopne	KbdWaitI6
KbdWaitI8:	retn
KbdWaitInpBuf	endp


;============================================================================

;		S u b r o u t i n e
; V: AL, CX, Flags

	; unused code >>
A20GateOn	proc near
		mov	ah,0DFh
		jmp	short A20GateO2
A20GateOn	endp


;============================================================================

;		S u b r o u t i n e
; V: AL, CX, Flags

A20GateOff	proc near		; bit 1 = gate A20
		mov	ah,0DDh
A20GateO2:	xor	cx,cx
A20GateO4:	in	al,64h		; AT Keyboard controller
		test	al,2
		loopne	A20GateO4
		mov	al,0D1h		; write output port.
					; The next byte written to port 60h will
					; be written to the 804x output port
		out	64h,al		; AT Keyboard controller
		xor	cx,cx
A20GateO6:	in	al,64h		; AT Keyboard controller
		test	al,2
		loopne	A20GateO6
		mov	al,ah
		out	60h,al		; AT Keyboard controller
		xor	cx,cx
A20GateO8:	in	al,64h		; AT Keyboard controller
		test	al,2
		loopne	A20GateO8
		retn
A20GateOff	endp
	; << unused code

flashp		ends

		end Start