; +--------------------------------------------------+
; | NexGen Flash BIOS update utility VL-Bus          |
; +--------------------------------------------------+

		title "NexGen Flash BIOS update utility VL-Bus"

; versions: (keep in sync with make.bat)
VL20		equ vers EQ 20
VL30		equ vers EQ 30

		.nolist
include biosdata.inc
include screen.inc
		.listall

StackSize       equ 200h
ImageAddress	equ 4000h		; in paragraphs
BiosAddress	equ 0e000h		; in paragraphs

EditBufSize	equ 40h

; ASCII character codes
BS		equ 8
LF              equ 10
CR              equ 13
ESCP            equ 27
DEL             equ 127

; Flash EPROM commands
FE_EraChip	equ 010h		; erase chip
FE_DPDisable	equ 020h		; software data protect disable
FE_EraSector	equ 030h		; erase sector
FE_Alternate	equ 060h		; alternate software product ID
FE_Prefix	equ 080h		; for 6 cycle commands
FE_Autoselect	equ 090h		; read device ID
FE_Program	equ 0A0h		; program byte(s)
FE_EraSecSusp	equ 0B0h		; during sector erase: suspend
					; resume via FE_EraSector
FE_Reset	equ 0F0h		; read/reset

; usage for Advanced Micro Devices (AMD) 29F010A:

;                            <- 1 -> <- 2 -> <- 3 -> <- 4 -> <- 5 -> <- 6 ->
;                        Cyc Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta
; Read (4)               1   RA  RD
; Reset (5)              1   xxx F0
; Reset (6)              3   555 AA  2AA 55  555 F0
; Autoselect (7):
; Manufacturer ID        4   555 AA  2AA 55  555 90  X00 01
; Device ID              4   555 AA  2AA 55  555 90  X01 20
; Sector Prot.Verify (8) 4   555 AA  2AA 55  555 90  SA+ 00/
;                                                    X02 01
; Program                4   555 AA  2AA 55  555 A0  PA  PD
; Chip Erase             6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 10
; Sector Erase           6   555 AA  2AA 55  555 80  555 AA  2AA 55  SA  30
; Erase Suspend (9)      1   XXX B0
; Erase Resume (10)      1   XXX 30

; usage for Silicon Storage Technology (SST) SST29EE010:

;                            <- 1 -> <- 2 -> <- 3 -> <- 4 -> <- 5 -> <- 6 ->
;                        Cyc Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta Adr Dta
; Software Data
; Protect Enable
; & Page Write         3+128 555 AA  2AA 55  555 A0  PA  PD ...
; Software Data
; Protect Disable        6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 20
; Software Chip Erase    6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 10
; Software ID Entry      3   555 AA  2AA 55  555 90
; Software ID Exit       3   555 AA  2AA 55  555 F0
; Alternate Software
; ID Entry               6   555 AA  2AA 55  555 80  555 AA  2AA 55  555 60


; useful macros
; define zero terminated string
dz              macro   n:req
		db n, 0
		endm

; combine x and y coordinates into word
yx              macro   y:req, x:req
		exitm   %x+(y*256)
		endm

; special NexGen opcode to control various undocumented processor features
nxop            macro
		db 0fh, 55h
		endm

flashv		segment public 'CODE'
		assume cs:flashv
		assume ds:nothing, es:nothing, ss:nothing

; +--------------------------------------------------+
; | VARIABLES and CONSTANTS                          |
; +--------------------------------------------------+

; program segment prefix (PSP)
		db 80h dup(?)		; unused
CmdLine         db 80h dup(?)

; program entry point
Start:		jmp	Start03
		cli			; =0FAh, unused

; v-- patch area for screen codes --v
TxUppLeftCornr:	dz ' '
TxUppRghtCornr:	dz ' '
TxLwrLeftCornr:	dz ' '
TxLwrRghtCornr:	dz ' '
TxHorizLine:	dz ' '
TxVertLine:	dz ' '
		dz ' '                  ; unused
		dz ' '                  ; unused
; ^-- patch area for screen codes --^

TxSpaceLine:	db 84 dup(' ')
		db 0

 IF VL20
TxBiosFilename:	db 'V06D033W.PRD'
 ELSE ;VL30
TxBiosFilename:	db 'V06E033W.PRD'
 ENDIF

TxSaveFilename:	db EditBufSize dup(0)
		db 100h dup(0)		; unused

Tx37NewLine: rept 37
		db LF, CR
	     endm
		db 0

StackArea:	dw StackSize dup(0)

		db 200h dup(0)          ; unused
		dw 0
DelayFact10	dw 0			; loops per 10 usec
DelayFact6	dw 0			; loops per  6 usec
		dw 0

TxSignon:	dz 'NEXGEN SYSTEM ROM BIOS UPDATE UTILITY'

TxDate:
 IF VL20
		dz '10/06/95'
 ELSE ;VL30
		dz '01/22/96'
 ENDIF

TxEnterUpdate:	dz 'Enter "FILENAME.XXX" to update in ROM, then press ENTER key.'

TxEnterSave:	dz 'Enter FILENAME.BIN to save modified BIOS, then press ENTER key.'

TxFileNotExist:	dz 'FILE DOES NOT EXIST, press any key to continue.'

TxFileSizeErr:	dz 'File size read error, press any key to continue.'

TxCantCreate:	dz 'Cannot create Filename, press any key to continue.'

TxPleaseWait:
 IF VL20
		db '       '
 ENDIF
		dz 'Please wait.  Reading the BIOS update information ...'

TxNotEnoughMem:	dz 'CANNOT CONTINUE, NOT ENOUGH LOW MEMORY AVAILABLE.'

TxContAbort:
 IF VL20
		db '               '
 ENDIF
		dz 'PRESS "C" TO CONTINUE'
		db 0

TxFileWriteErr:	dz 'File write error, press any key to continue.'

TxChecksumErr:	dz 'FILE HAS BIOS CHECKSUM ERROR, Press any key to continue'

TxDateIdentic:	db 'CANNOT CONTINUE, FILE AND ROM DATA ARE IDENTICAL, '
		dz 'press a key to restart.'

TxNoNexGenBios:	dz 'FILE IS NOT A VALID NEXGEN BIOS, Press any key to continue'

TxWrongSystem:	dz 'YOU CANNOT USE THIS UTILITY ON THIS SYSTEM, Press any key to continue'

TxUpdComplete:	dz 'BIOS UPDATE COMPLETE'

TxWriteProtect:	dz 'Cannot continue ..., FLASH MEMORY is write protected.'

TxBadFlashRom:	dz 'Cannot continue ..., unidentified or bad FLASH MEMORY.'

TsSuccess:	dz '                       UPDATE COMPLETE.'
		dz ' '
 IF VL20
		db '       '
 ENDIF
		dz 'TURN OFF THE COMPUTER, THEN BACK ON, TO ACTIVATE NEW BIOS.'
		db 0

TsFinalWarning:	dz '                 WARNING !!!'
		dz ' '
		dz 'THIS IS YOUR LAST CHANCE TO DISCONTINUE THE UPDATE.'
		dz 'TO ABORT THE UPDATE PRESS "ESCAPE" NOW.'
		dz 'IF YOU CONTINUE, THE UPDATE MAY TAKE UP TO 30'
		dz 'SECONDS.  DO NOT RESET OR POWER DOWN THE SYSTEM'
		dz 'WHILE UPDATING.'
		db 0

TsPleaseNote:	dz 'PLEASE NOTE:'
		dz ' '
		dz '----------- '
		dz ' '
		dz 'This utility program may be affected by other drivers loaded'
		dz 'in memory.  If you encounter any problems during the update'
		dz 'process, you should reset the system and do a clean boot.'
		dz ' '
		dz '(See the README.TXT for directions on how to do a clean boot.'
		dz ' If you want to read the file NOW, press "ESCAPE" key to abort'
 IF VL20
		dz ' this program.  Then, enter "TYPE README.TXT|MORE")'
 ELSE ;VL30
		dz ' this program.  Then, enter "TYPE README.TXT|MORE".)'
 ENDIF
		dz ' '
		dz 'After the clean boot, run this update utility again.'
		dz ' '
		dz '               Press "C" to continue ...'
 IF VL30
		dz ' '
		dz '(This info is also available on NexGen Online at '
		dz ' http://www.nexgen.com)'
		dz ' '
 ENDIF
		db 0

TxAbortUpdate:	dz 'ABORT BIOS UPDATE.'
TxBeingUpdated:
 IF VL20
		db '       '
 ENDIF
		dz 'UPDATING THE BIOS'
TsDoNotReset:	dz 'DO NOT RESET OR TURN-OFF SYSTEM.'
		dz 'UPDATING MAY TAKE UP TO 30 SECONDS.'
		dz 'WAIT FOR COMPLETION MESSAGE.'
		db 0

FileHandle1     dw 0
FileHandle2     dw 0
CharAttr        db BG_Blue+FG_White
		db BG_Blue+FG_White

TsCursorKeys:	dz 'Insert key -> to insert a character      Left cursor key --> to go left'
		dz 'Delete key -> to delete a character      Right cursor key -> to go right'
		dz '                              ESC --> To abort'
		db 0

TxControlKeys:	dz 'Control Keys'

TsThankYou:
 IF VL20
		db '               '
 ENDIF
		dz 'Thank you for choosing NexGen.'
		dz ' '
		dz 'You have activated the NexGen BIOS flash utility for your'
		dz 'Nx586 VL-based PC.  This utility will update your BIOS'
		dz 'quickly and effectively.'
		dz ' '
 IF VL20
		db '               '
 ENDIF
		dz 'Press any key to continue ...'
		db 0

 IF VL30
TsPciBiosOnVl:	dz 'You are attempting to update a PCI system BIOS with a VL'
		dz 'update utility.  You must use the PCI version.  This utility'
		dz 'is part of the update kit: UPDTPIxx.EXE (where "xx"'
		dz 'is the version number).  It can be obtained from NexGen Online:'
		dz '               BBS - 408.955.1839'
		dz '               WEB - http://www.nexgen.com'
		dz ' '
		dz 'Press any key ...'
		db 0
 ENDIF

TxDot:		dz '.'

ImageSegment	dw ImageAddress


; +--------------------------------------------------+
; | CODE                                             |
; +--------------------------------------------------+

; Scan command line of PSP
; E: DS = ES = PSP, CS = SS = code segment

Start03:	cld
		cli
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	ss,ax
		mov	sp,offset StackArea+2*StackSize
 IF VL30
		mov	si,0B000h
		nxop			; get BIOS version
		cmp	cl,3		; is it VL system?
		jz	Main03		; yes, continue
		; Nx586 PCI system - tell user and exit
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	bx,offset TsPciBiosOnVl
		mov	ax,yx(4,7)
		call	PrintPosMulti
		call	GetKeyb
		jmp	TerminateProg

Main03:		; clear data area
 ENDIF
		mov	di,offset TxSaveFilename
		xor	ax,ax
		mov	cx,(EditBufSize+100h)/2
	   repe	stosw
		mov	si,offset Tx37NewLine
		call	Print		; a simple form of ClearScreen

		; no help, check memory requirements
		mov	ax,cs		; check if end of program below image address
		cmp	ax,ImageAddress-500h ; reserve 20KB for program
		jbe	Main12

		; not enough memory - tell user and exit
		mov	bx,offset TxNotEnoughMem
		mov	ax,yx(8,7)
		call	PrintPos
		jmp	PrintTermProg

Main12:
Main18:		; clear image area
		cld
		push	es
		mov	ax,cs:ImageSegment
		mov	es,ax
		xor	di,di
		xor	ax,ax
		mov	cx,8000h	; fill 32K words with 0
	    rep stosw			; fill CX times AX to ES:DI
		mov	ax,es
		add	ax,1000h
		mov	es,ax
		xor	di,di
		xor	ax,ax
		mov	cx,8000h	; fill 32K words with 0
	    rep stosw			; fill CX times AX to ES:DI
		pop	es
		mov	bx,offset TxBiosFilename+12

Main21:		cmp	byte ptr [bx],0 ; advance to terminating 0
		jz	Main24
		inc	bx
		jmp	short Main21

Main24:		dec	bx		; check char before 0
		cmp	byte ptr [bx],' '
		jz	Main24
		inc	bx
		mov	byte ptr [bx],0	; new end of string
		; enter editor
		push	bx
		mov	bx,offset TxSaveFilename
		mov	ax,yx(5,7)
		call	PrintPos
		pop	bx
		mov	si,offset TxSaveFilename-1
		mov	di,offset TxSaveFilename+40h-1
		mov	dx,offset TxBiosFilename
		call	OpenFile
		jnb	Main30		; no error
		jmp	EFileNotExist	; open failed

Main30:		push	ax
		; show first screen
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	ax,yx(3,7)
		mov	bx,offset TsThankYou
		call	PrintPosMulti
		call	GetKeyb

		; show second screen
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,77)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	ax,yx(3,7)
		mov	bx,offset TsPleaseNote
		call	PrintPosMulti
		call	ContinueAbort
		pop	ax		; filehandle
		mov	cs:FileHandle1,ax
		push	ax

		; show third screen
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	bx,offset TxPleaseWait
		mov	ax,yx(9,7)
		call	PrintPos

		; file is open, filehandle on stack
		; read BIOS file (128K) into memory
		pop	ax
		mov	bx,ax
		xor	dx,dx
		call	Read32KFromFile
		push	ax
		mov	ax,yx(11,30)
		call	PrintDot
		pop	ax
		mov	bx,cs:FileHandle1
		mov	dx,8000h
		call	Read32KFromFile
		push	ax		; not necessary
		mov	ax,yx(11,32)
		call	PrintDot
		pop	ax
		mov	ax,cs:ImageSegment
		add	ax,1000h
		mov	cs:ImageSegment,ax
		xor	dx,dx
		call	Read32KFromFile
		push	ax
		mov	ax,yx(11,34)
		call	PrintDot
		pop	ax
		mov	bx,cs:FileHandle1
		mov	dx,8000h
		call	Read32KFromFile
		push	ax		; not necessary
		mov	ax,yx(11,36)
		call	PrintDot
		pop	ax
		mov	ax,cs:ImageSegment
		sub	ax,1000h
		mov	cs:ImageSegment,ax

		; Check for magic 1234h at 4012h (corresponds to address 0E4012h).
		; This works for VL-Bios only, PCI-Bios has magic at 0E0012h.
		push	ds
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,ds:4012h
		pop	ds
		cmp	ax,1234h	; some magic number
		jz	Main33		; no effect
		jmp	ENoNexGenBios

Main33:		; Check Checksum. This too is nonsense: result is discarded ..
		; Read word at offset 0FFFAh (corresponds to address 0EFFFAh)
		push	ds
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,ds:0FFFAh
		pop	ds
		cmp	ax,0		; no checksum?
		jz	Main36		; then ignore it
		; checksum present, check it
		call	CheckChecksum
		jnb	Main36		; no effect
		mov	bx,offset TxChecksumErr
		jmp	EClosePrTerm

Main36:		cli
		; adapt jumps in code
		mov	bx,offset FarJumpProt+1
		mov	ax,bx
		add	ax,4
		mov	cs:[bx],ax
		mov	bx,offset FarJumpReal+1
		mov	ax,bx
		add	ax,4
		mov	cs:[bx],ax
		add	bx,2
		mov	ax,cs
		mov	cs:[bx],ax
		; Compare BIOS image with current BIOS
		; Another nonsense: this compare will always fail
		; because, to support BIOS calls, the shadow RAM
		; image is currently mapped, not the EPROM image
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ax,0E000h
		mov	es,ax
		mov	cx,8000h	; 32K words
		xor	si,si
		xor	di,di
Main39:		lodsw
		scasw
		jnz	Main45
		loop	Main39
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
		mov	ax,0F000h
		mov	es,ax
		mov	cx,8000h
		xor	si,si
		xor	di,di
Main42:		lodsw
		scasw
		jnz	Main45
		loop	Main42
		; BIOS identical
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		sti
		mov	ax,yx(12,64)
		mov	bx,offset TxDateIdentic
		call	PrintPosMulti
		call	GetKeyb
		jmp	EClosePrTerm

Main45:		; BIOS different
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		sti
		mov	bx,cs:FileHandle1
		call	CloseFile
		mov	si,0B001h	; switch off caches (1st and 2nd)
		mov	al,0
		nxop
		; calibrate delay loops
		; uses BIOS data timer_low, which is incremented every
		; 55 ms (18.2 x per second) by the timer interrupt routine
		push	ds
		mov	ax,bdata
		mov	ds,ax
		assume ds:bdata
		mov	si,1000h
Calibr03:	mov	dl,16
		mov	cx,0
		; synchronise with timer tick for accurate measurement
		mov	ax,timer_low
Calibr06:	cmp	timer_low,ax
		jz	Calibr06
		; now time 16 x 64K loops
		mov	ax,timer_low
Calibr09:	loop	Calibr09	; loop 64K times
		dec	dl
		jnz	Calibr09
		mov	bx,timer_low
		cmp	bx,ax		; timer wrap around?
		jbe	Calibr03	; yes, repeat measurement
		sub	bx,ax		; BX = ticks (55 ms) per 16*64K loops
		mov	ax,si
		xor	dx,dx
		div	bx		; divide 1000h by ticks
		xor	dx,dx
		mov	bx,55		; msec per tick
		div	bx
		; AX = loops per 1/256 msec
		; Shift AX left 0, 4 or 8 before division by 100 for
		; maximum accuracy, then shift left again 8, 4 or 0
		; i.e. multiply by 256
		mov	cx,0800h
		cmp	ax,1000h	; faster than 1000h loops per 1/256 msec?
		jnb	Calibr15	; yes
		mov	cx,0404h
		cmp	ax,0100h	; faster than 100h loops per 1/256 msec?
		jnb	Calibr12	; yes
		mov	cx,0008h	; no, slower
Calibr12:	shl	ax,cl		; shift by 8, 4 or not
Calibr15:	push	ax
		xor	dx,dx
		mov	bx,100
		div	bx
		mov	cl,ch
		or	cl,cl
		jz	Calibr18
		shl	ax,cl
Calibr18:	; AX = loops per 1/100 msec (10 usec)
		mov	cs:DelayFact10,ax
		pop	ax
		xor	dx,dx
		mov	bx,167
		div	bx
		mov	cl,ch
		or	cl,cl
		jz	Calibr21
		shl	ax,cl
Calibr21:	; AX = loops per 1/167 msec (6 usec)
		mov	cs:DelayFact6,ax
		pop	ds

		; now flash begins
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	ax,yx(0,20)
		mov	bx,offset TxSignon
		call	PrintPos
		mov	ax,yx(1,32)
		mov	bx,offset TxDate
		call	PrintPos
		mov	ax,yx(15,13)
		mov	bx,offset TxContAbort
		call	PrintPos
		mov	al,BG_Blue+FG_LtRed
		call	SetCharAttr
		mov	ax,yx(5,13)
		mov	bx,offset TsFinalWarning
		call	PrintPosMulti
		call	ContinueAbort

		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	al,BG_BlBlack+FG_BrWhite
		call	SetCharAttr
		mov	ax,yx(8,21)
		mov	bx,offset TxBeingUpdated
		call	PrintPos
		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(10,21)
		mov	bx,offset TsDoNotReset
		call	PrintPosMulti
		xor	bp,bp
		cli
		mov	al,1		; BIOS shadow RAM image is swapped now with BIOS EPROM image
		call	near ptr ProgChipSet
		; detect Flash EPROM device
		mov	ax,0E000h
		mov	es,ax
		call	EnableFlashWrite
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Autoselect
		call	DisableFlashWrite
		mov	al,es:0		; read manufacturer ID
		; 01 = Advanced Micro Devices, AMD
		cmp	al,01h		; AMD ?
		jnz	FlashProg9
		mov	al,es:1
		; 20 = Am29F010(A) 128K x 8 Sector Erase Flash Memory
		cmp	al,20h		; Am29F010(A) ?
		jnz	FlashProg9
		jmp	FlashAMD

FlashProg9:	; not Am29F010(A)
		call	EnableFlashWrite
		mov	byte ptr es:5555h,0AAh	; alternate software ID entry
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Alternate
		call	Wait10m
		mov	al,es:0		; read manufacturer and device ID
		mov	ah,es:1
		; BF = Silicon Storage Technology, SST
		; 07 = SST29EE010 128K x 8 Page Mode EEPROM
		cmp	ax,07BFh	; SST 29EE010 ?
		jz	FlashSST
		jmp	EBadFlash	; unknown device

FlashSST:	; SST29EE010 128K x 8 Page Mode EEPROM
		; see http://www.ssti.com/products/29xe010.html
		mov	byte ptr es:5555h,0AAh	; software ID exit code
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Reset
		call	Wait10m
		mov	byte ptr es:5555h,0AAh	; software data protect disable
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_DPDisable
		call	Wait10m
		mov	byte ptr es:5555h,0AAh	; software chip erase
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_EraChip
		call	Wait10m
		cld
		mov	ax,cs:ImageSegment
		mov	ds,ax

FlashSST03:	; program 64 KB per loop
		xor	si,si
		xor	di,di

FlashSST06:	; program 128 byte per loop
		mov	cx,80h
	    rep movsb				; copy 128 byte
		call	Wait300u		; T_BLCO
		dec	si
		dec	di
		lodsb				; get last written byte from image

FlashSST12:	mov	ah,es:[di]		; get status byte from flash chip
		cmp	al,ah			; same == done?
		jnz	FlashSST12		; no - check again
		inc	di			; to next 128 byte
		or	di,di			; 64 KB done?
		jnz	FlashSST06		; no - continue
		; 64 KB done
		mov	ax,es
		cmp	ax,0F000h		; did we just do the second 64 KB?
		jz	FlashSST15		; yes - done
		add	ax,1000h		; no, advance segment
		mov	es,ax
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
 IF VL20
		pop	bx
 ENDIF
		jmp	short FlashSST03	; next (final) 64 KB

FlashSST15:	; done
 IF VL20
		pop	bx
 ENDIF
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Program
		call	Wait10m
		call	DisableFlashWrite
		jmp	short FlashComp

FlashAMD:	; Am29F010(A) 128K x 8 Sector Erase Flash Memory
		; see http://www.amd.com/products/nvd/techdocs/techdocs.html#29Fseries
		; ensure that all sectors are writeable, then erase and program chip
		mov	cx,8
		mov	bx,2

FlashAMD03:	call	EnableFlashWrite
		or	bx,2
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Autoselect
		call	DisableFlashWrite
		mov	al,es:[bx]
		cmp	al,1		; sector protected?
		jz	EWriteProtect	; write protect error
		cmp	al,0
		jnz	EBadFlash	; problem
		and	bx,0F000h
		add	bx,4000h
		or	bx,bx
		jnz	FlashAMD12
		mov	ax,0F000h
		mov	es,ax
FlashAMD12:	loop	FlashAMD03

FlashAMD15:	mov	ax,cs:ImageSegment
		mov	ds,ax
		call	FProgAMD
		jb	FlashAMD15
		xor	cx,cx
		loop	$		; wait a bit

FlashComp:	; programming Am29F010 or SST29EE010 complete
		mov	bx,0

FlashComp3:	mov	ds,bx
		xor	si,si
		xor	di,di
		mov	cx,8000h
	    rep lodsw
		add	bh,10h
		cmp	bh,0A0h
		jnz	FlashComp3
		; flash program went ok, now halt machine (wait for power off)
		call	MapShadow
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	al,BG_Black+FG_White
		call	SetCharAttr
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		mov	bx,offset TsSuccess
		mov	ax,yx(6,2)
		call	PrintPosMulti
		jmp	short $

EBadFlash:	; programming problem
		call	MapShadow
		mov	bx,offset TxBadFlashRom
		jmp	PrintTermProg

EWriteProtect:	; write protect
		call	MapShadow
		mov	bx,offset TxWriteProtect
		jmp	PrintTermProg

;============================================================================

;		S u b r o u t i n e

MapShadow	proc near
		mov	al, 4
		call	ProgChipSet
		mov	si,0B001h
		mov	al,1
		nxop
		call	DisableFlashWrite
		retn
MapShadow	endp


;============================================================================

;		S u b r o u t i n e
; Wait 300 usec (30 x 10 usec)

Wait300u	proc near
		push	cx
		mov	cx,30
Wait300u3:	call	Wait10u
		loop	Wait300u3
		pop	cx
		retn
Wait300u	endp


;============================================================================

;		S u b r o u t i n e
; Wait 6 usec

	; unused code >>
Wait6u		proc near
		push	cx
		mov	cx,cs:DelayFact6
		loop	$
		pop	cx
		retn
Wait6u		endp
	; << unused code


;============================================================================

;		S u b r o u t i n e
; Wait 10 usec

Wait10u		proc near
		push	cx
		mov	cx,cs:DelayFact10
		loop	$
		pop	cx
		retn
Wait10u		endp


;============================================================================

;		S u b r o u t i n e
; Wait 10 msec (1000 x 10 usec)

Wait10m		proc near
		push	cx
		push	ax
		mov	ax,1000
Wait10m3:	mov	cx,cs:DelayFact10
		loop	$
		dec	ax
		jnz	Wait10m3
		pop	ax
		pop	cx
		retn
Wait10m		endp


;============================================================================

;		S u b r o u t i n e

FProgAMD	proc near
		call	EnableFlashWrite
		mov	ax,0E000h
		mov	es,ax
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Prefix
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_EraChip
		call	DisableFlashWrite

FProgAMD03:	mov	al,es:0
		test	al,80h
		jnz	FProgAMD06
		test	al,20h
		jnz	FProgAMD
		jmp	short FProgAMD03

FProgAMD06:	mov	ax,0E000h
		mov	es,ax
		cld

FProgAMD09:	mov	cx,0
		xor	bx,bx
		xor	si,si

FProgAMD12:	lodsb
		mov	ah,al
		call	EnableFlashWrite
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Program
		mov	es:[bx],ah
		call	DisableFlashWrite
		mov	dh,ah
		and	dh,80h

FProgAMD15:	mov	al,es:[bx]
		and	al,80h
		cmp	al,dh
		jnz	FProgAMD15
		mov	al,es:[bx]
		cmp	al,ah
		jnz	FProgAMD15
		inc	bx
		loop	FProgAMD12
		mov	ax,es
		cmp	ax,0F000h
		jz	FProgAMD18
		add	ax,1000h
		mov	es,ax
		mov	ax,ds
		add	ax,1000h
		mov	ds,ax
		jmp	short FProgAMD09

FProgAMD18:	dec	bx
		call	EnableFlashWrite
		mov	byte ptr es:5555h,0AAh
		mov	byte ptr es:2AAAh,55h
		mov	byte ptr es:5555h,FE_Reset
		mov	al,es:[bx]
		clc
		retn
FProgAMD	endp


;============================================================================

;		S u b r o u t i n e

; Enable flash EPROM programming
EnableFlashWrite proc near
		push	ax
		push	bx
		push	cx
		push	dx
		push	si
		push	es
		push	ds
		mov	al,2
		call	ProgChipSet
		pop	ds
		pop	es
		pop	si
		pop	dx
		pop	cx
		pop	bx
		pop	ax
		retn
EnableFlashWrite endp


;============================================================================

;		S u b r o u t i n e

; Disable flash EPROM programming
DisableFlashWrite proc near
		push	ax
		push	bx
		push	cx
		push	dx
		push	si
		push	es
		push	ds
		mov	al,3
		call	ProgChipSet
		pop	ds
		pop	es
		pop	si
		pop	dx
		pop	cx
		pop	bx
		pop	ax
		retn
DisableFlashWrite endp


;============================================================================

;		S u b r o u t i n e

Read32KFromFile	proc near
		mov	cx,8000h
		push	ds
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	ah,3Fh
		int	21h		; DOS - 2+ - READ FROM FILE WITH HANDLE
					; BX = file handle, CX = number of bytes to read
					; DS:DX -> buffer
		pop	ds
		retn
Read32KFromFile	endp

	; unused code >>
		jnb	Read32Trash
		pop	ax
		jmp	short EFileSize

Read32Trash:	cmp	ax,8000h
		jz	Read32Trash3
		pop	ax
		jmp	short EFileSize

Read32Trash3:	retn
	; << unused code

;============================================================================

;		S u b r o u t i n e
; Add all words of 128K BIOS image
; E: ImageSegment
; R: Carry if check failed, NCarry if check ok

CheckChecksum	proc near
		push	ds
		cld
		mov	ax,cs:ImageSegment
		mov	ds,ax
		mov	cx,8000h
		xor	bx,bx
		xor	ax,ax
ChkChksum3:	add	ax,[bx]
		add	bx,2
		loop	ChkChksum3
		mov	bx,ds
		add	bx,1000h
		mov	ds,bx
		mov	cx,8000h
		xor	bx,bx
ChkChksum6:	add	ax,[bx]
		add	bx,2
		loop	ChkChksum6
		pop	ds
		cmp	ax,0		; ok?
		jnz	ChkChksum9	; no
		; check ok
		clc
		retn

ChkChksum9:	; check failed
		stc
		retn
CheckChecksum	endp

		assume ds:flashv

	; unused code >>
		mov	bx,FileHandle1
		call	CloseFile
		mov	bx,offset TxWrongSystem
		jmp	short EPrintTerm
	; << unused code

ENoNexGenBios:	mov	bx,FileHandle1
		call	CloseFile
		mov	bx,offset TxNoNexGenBios
		jmp	short EPrintTerm

EFileSize:	mov	bx,FileHandle1
		call	CloseFile
		mov	bx,offset TxFileSizeErr
		jmp	short EPrintTerm

EFileNotExist:	mov	bx,offset TxFileNotExist
EPrintTerm:	; print error message and try again
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		mov	ax,yx(8,7)
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		call	GetKeyb
		jmp	Main12

EClosePrTerm:	; close file, print error message and try again
		push	bx
		mov	bx,FileHandle1
		call	CloseFile
		pop	bx
		jmp	short EPrintTerm

	; unused code >>
AskDest:	; ask for new destination filename
		mov	ax,cs
		mov	ds,ax
		mov	ax,yx(13,1)	; from
		mov	cx,yx(18,78)	; to
		call	ClearScreenArea
		mov	al,BG_Blue+FG_Red
		call	SetCharAttr
		mov	ax,yx(14,6)	; from
		mov	cx,yx(16,70)	; to
		call	ClearScreenArea
		mov	ax,yx(13,7)
		mov	bx,offset TxEnterSave
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		; cut trailing spaces from TxBiosFilename
		mov	bx,offset TxBiosFilename
AskDest2:	cmp	byte ptr [bx],0
		jz	AskDest4
		inc	bx
		jmp	short AskDest2

AskDest4:	dec	bx
		cmp	byte ptr [bx],' '
		jz	AskDest4
		inc	bx
		mov	byte ptr [bx],0
		; get new filename from user
		push	bx
		mov	bx,offset TxBiosFilename
		mov	ax,yx(15,7)
		call	PrintPos
		pop	bx
		mov	si,offset TxBiosFilename-1
		mov	di,offset TxBiosFilename-1+EditBufSize
		call	Editor
		; try to make backup
		mov	dx,offset TxBiosFilename
		call	CreateFile
		jb	ECantCreate
		mov	FileHandle2,ax
		mov	bx,ax
		mov	ax,es:ImageSegment
		mov	ds,ax
		mov	cx,0FFFFh
		mov	dx,0
		call	WriteFile
		jb	EFileWrite
		call	FilePosFFFF
		mov	dx,0FFFFh
		mov	cx,1
		mov	bx,es:FileHandle2
		call	WriteFile
		jb	EFileWrite
		mov	bx,offset TxUpdComplete
	; << unused code

PrintTermProg:	; output message and terminate program
		push	bx
		mov	ax,cs
		mov	ds,ax
		mov	es,ax
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		pop	bx
		mov	ax,yx(6,16)
		call	PrintPos
		mov	dx,yx(24,0)
		call	SetCursor
		jmp	short TerminateProg

	; unused code >>
ECantCreate:	mov	bx,offset TxCantCreate
		jmp	short EBackup

EFileWrite:	mov	bx,FileHandle2
		call	CloseFile
		mov	bx,offset TxFileWriteErr

EBackup:	; problem with backup file
		; ask user for a new destination filename
		mov	al,BG_Blue+FG_Magenta
		call	SetCharAttr
		mov	ax,yx(17,7)
		call	PrintPos
		mov	al,BG_Blue+FG_White
		call	SetCharAttr
		call	GetKeyb
		jmp	AskDest
	; << unused code

TerminateProg:	sti
		mov	ax,4C01h	; QUIT WITH EXIT CODE (AL = code)
		int	21h		; DOS
		jmp	short $		; halt system (never reached)

;============================================================================

;		S u b r o u t i n e

OpenFile	proc near
		mov	ah,3Dh
		mov	al,0
		int	21h		; DOS - 2+ - OPEN DISK FILE WITH HANDLE
					; DS:DX -> ASCIZ filename
					; AL = access mode
					; 0 - read
		retn
OpenFile	endp


;============================================================================

;		S u b r o u t i n e

CloseFile	proc near
		mov	ah,3Eh
		int	21h		; DOS - 2+ - CLOSE A FILE WITH HANDLE
					; BX = file handle
		retn
CloseFile	endp


;============================================================================

;		S u b r o u t i n e

CreateFile	proc near
		mov	ah,3Ch
		xor	cx,cx
		int	21h		; DOS - 2+ - CREATE A FILE WITH HANDLE (CREAT)
					; CX = attributes for file
					; DS:DX -> ASCIZ filename (may include drive and path)
		retn
CreateFile	endp


;============================================================================

;		S u b r o u t i n e

WriteFile	proc near
		mov	ah,40h
		int	21h		; DOS - 2+ - WRITE TO FILE WITH HANDLE
					; BX = file handle, CX = number of bytes to write, DS:DX -> buffer
		retn
WriteFile	endp


;============================================================================

;		S u b r o u t i n e

FilePosFFFF	proc near
		mov	ax,4200h
		mov	cx,0
		mov	dx,0FFFFh
		int	21h		; DOS - 2+ - MOVE FILE READ/WRITE POINTER (LSEEK)
					; AL = method: offset from beginning of file
		retn
FilePosFFFF	endp


;============================================================================

;		S u b r o u t i n e
; E: pointer into buffer in BX, start of buffer in SI, end of buffer in DI
; R: buffer (no return if ESC key pressed)

	; unused code >>
Editor03:	inc	bx		; undo dec

Editor		proc near
		call	GetKeyb
		cmp	al,CR		; done?
		jnz	Editor06	; no, edit
		retn
Editor		endp
	; << unused code


;============================================================================

;		S u b r o u t i n e
; E:-
; R:- (No return if ESC key pressed)

ContAbor6:	; handle abort
		pop	ax		; drop return address
		mov	bx,offset TxAbortUpdate
		jmp	short PrintTermProg

ContinueAbort	proc near
		call	GetKeyb
		and	al,0DFh		; convert to upper case
		cmp	al,'C'		; continue?
		jnz	ContAbor3
		retn

ContAbor3:	cmp	al,ESCP		; abort?
		jnz	ContinueAbort
		jmp	short ContAbor6
ContinueAbort	endp


;============================================================================

; extension for subroutine Editor
; AL= ascii char, AH= scan code
; BX= pointer into buffer
; SI= start of buffer, DI= end of buffer

	; unused code >>
Editor06:	cmp	al,ESCP		; abort?
		jnz	Editor09	; no
		; handle abort
		mov	ax,cs
		mov	ds,ax
		mov	ax,yx(0,0)	; from
		mov	cx,yx(23,79)	; to
		call	ClearScreenArea
		pop	ax		; drop return address
		jmp	short TerminateProg

Editor09:	cmp	ah,53h		; delete key?
		jz	Editor30
		cmp	ah,52h		; insert key?
		jnz	Editor12
		jmp	Editor48

Editor12:	cmp	al,BS		; backspace
		jz	Editor18
		cmp	ah,4Bh		; cursor left key?
		jz	Editor18
		cmp	ah,4Dh		; cursor right key?
		jz	Editor21
		cmp	al,' '		; control char?
		jb	Editor	; ignore
		cmp	al,'@'		; < A-Za-z?
		jb	Editor15
		cmp	al,DEL		; > DEL?
		ja	Editor	; ignore

Editor15:	; put char into buffer and echo it
		cmp	bx,di		; end of buffer reached?
		jz	Editor	; yes, ignore
		mov	[bx],al		; no, put in
		mov	es:TxTmpChar,al
		push	si
		mov	si,offset TxTmpChar ; print char
		call	PrintBS
		pop	si
		inc	bx
		jmp	short Editor

Editor18:	; handle backspace / cursor left
		dec	bx
		cmp	bx,si		; start of buffer reached?
		jz	Editor03
		push	si
		mov	si,offset TxBackSpace
		call	PrintBS		; print BS
		pop	si
		jmp	short Editor

Editor21:	; handle cursor right
		mov	al,[bx]
		cmp	al,0		; if at end of string
		jnz	Editor24
		mov	al,' '		; enlarge string with space
Editor24:	mov	es:TxTmpChar,al
		inc	bx
		cmp	bx,di		; end of buffer reached?
		jz	Editor27
		push	si
		mov	si,offset TxTmpChar
		call	PrintBS		; print char
		pop	si
		jmp	Editor

Editor27:	dec	bx		; undo inc
		jmp	Editor

Editor30:	; handle delete
		cmp	bx,di		; at end of buffer?
		jnz	Editor33
		jmp	short Editor45

Editor33:	push	bx
Editor36:	inc	bx
		cmp	bx,di
		jz	Editor39
		mov	al,[bx]
		dec	bx
		mov	[bx],al
		inc	bx
		cmp	bx,di
		jz	Editor39
		cmp	al,0
		jnz	Editor36
		dec	bx
		mov	byte ptr [bx],' '
		inc	bx
		mov	byte ptr [bx],' '
		jmp	short Editor42

Editor39:	dec	bx
		mov	byte ptr [bx],' '
Editor42:	pop	bx
		push	bx
		push	si
		mov	si,bx
		call	GetCursor
		push	dx
		call	Print
		pop	dx
		call	SetCursor	; E: xy in DX
		pop	si
		pop	bx
Editor45:	jmp	Editor

Editor48:	; handle insert
		cmp	bx,di		; at end of buffer?
		jz	Editor45
		push	bx
		mov	ah,' '
		mov	al,[bx]
		mov	[bx],ah
Editor51:	inc	bx
		cmp	bx,di
		jz	Editor54
		mov	ah,[bx]
		mov	[bx],al
		cmp	al,0
		jz	Editor54
		inc	bx
		cmp	bx,di
		jz	Editor54
		mov	al,[bx]
		mov	[bx],ah
		cmp	ah,0
		jz	Editor54
		jmp	short Editor51

Editor54:	pop	bx
		push	bx
		push	si
		mov	si,bx
		call	GetCursor
		push	dx
		call	Print
		pop	dx
		call	SetCursor	; E: xy in DX
		pop	si
		pop	bx
		jmp	Editor

TxTmpChar	db 0,0			; temporary storage to print one char
TxBackSpace	db BS,0
	; << unused code

;============================================================================

;		S u b r o u t i n e
; E: xy in AX

SetCursorSav	proc near
		push	ax
		push	bx
		mov	dx,ax
		call	SetCursor	; E: xy in DX
		pop	bx
		pop	ax
		retn
SetCursorSav	endp


;============================================================================

;		S u b r o u t i n e
; E: xy in DX

SetCursor	proc near
		mov	ah,2
		mov	bh,0
		int	10h		; - VIDEO - SET CURSOR POSITION
					; DH,DL = row, column (0,0 = upper left)
					; BH = page number
		retn
SetCursor	endp


;============================================================================

;		S u b r o u t i n e

GetCursor	proc near
		mov	ah,3
		mov	bh,0
		int	10h		; - VIDEO - READ CURSOR POSITION
					; BH = page number
					; Return: DH,DL = row,column, CH = cursor start line, CL = cursor end line
		retn
GetCursor	endp


;============================================================================

;		S u b r o u t i n e
; Another print routine. Sigh!
; Prints at the current cursor position. This time it handles backspaces too.
; E: pointer to string in SI
; R: SI points after terminating 0, AH=14

PrintBS		proc near
		push	di
		push	bx
		xor	bh,bh
PrintBS3:	mov	ah,14		; why??
		lods	byte ptr es:[si]
		test	al,al		; =0?
		jz	PrintBS9
		cmp	al,BS		; backspace?
		jnz	PrintBS6	; no
		; handle backspace
		call	GetCursor
		dec	dx		; back one X-position
		call	SetCursor
		jmp	short PrintBS3

PrintBS6:	; print char in AL
		push	ax
		call	GetCursor
		pop	ax
		push	dx
		push	bx
		mov	bh,0		; display page
		mov	bl,BG_Blue+FG_White
		mov	cx,1		; number of times to write char
		mov	ah,9		; WRITE ATTRIBUTES/CHARACTERS AT CURSOR POSITION
		int	10h		; VIDEO
		pop	bx
		pop	dx
		inc	dx		; forward one X-position
		call	SetCursor
		jmp	short PrintBS3

PrintBS9:	pop	bx
		pop	di
		retn
PrintBS		endp


;============================================================================

;		S u b r o u t i n e
; E: pointer to string in SI
; R: SI points after terminating 0
Print		proc near
		push	di
		push	bx
		xor	bh,bh		; display page 0

Print3:		mov	ah,14		; WRITE CHARACTER AND ADVANCE CURSOR (TTY WRITE)
		lodsb
		test	al,al		; terminating 0?
		jz	Print6		; yes, done
		int	10h		; VIDEO
					; AL = character, BH = display page (alpha modes)
					; BL = foreground color (graphics modes)
		jmp	short Print3

Print6:		pop	bx
		pop	di
		retn
Print		endp


;============================================================================

;		S u b r o u t i n e
; E: xy in AX, pointer to string in BX
; R: BX points to terminating 0

PrintPos	proc near
		push	ax
		push	bx
		push	dx
		push	cx
		push	di
		call	SetCursorSav	; E: xy in AX
PrintPo2:	push	ax
		mov	al,[bx]
		cmp	al,0
		jz	PrintPo4
		push	bx
		mov	bh,0
		mov	bl,es:CharAttr
		mov	cx,1
		mov	ah,9		; WRITE ATTRIBUTES/CHARACTERS AT CURSOR POSITION
		int	10h		; VIDEO
					; AL = character, BH = display page
					; BL = attributes of character (alpha modes) or color (graphics modes)
					; CX = number of times to write character
		pop	bx
		pop	ax
		inc	ax
		call	SetCursorSav	; E: xy in AX
		inc	bx
		jmp	short PrintPo2

PrintPo4:	pop	ax
		pop	di
		pop	cx
		pop	dx
		pop	bx
		pop	ax
		retn
PrintPos	endp


;============================================================================

;		S u b r o u t i n e
; Print a list of strings, each one at a new line
; E: xy in AX, pointer to string in BX

PrintPosMulti	proc near
		call	PrintPos
		inc	ah		; increment Y-position
PrintPosM2:	inc	bx		; skip terminating 0
		cmp	byte ptr cs:[bx],0
		jnz	PrintPosM2
		inc	bx		; goto next string
		cmp	byte ptr cs:[bx],0 ; done?
		jnz	PrintPosMulti	; no, print it
		retn
PrintPosMulti	endp

	; unused code >>
		call	PrintRectangle  ; E: xy upper left corner in AX, xy lower right corner in CX
		retn
	; << unused code

;============================================================================

;		S u b r o u t i n e
; E: xy upper left corner in AX, xy lower right corner in CX

PrintRectangle	proc near
		push	ax
		push	cx
		inc	al
		sub	cl,al
		xor	ch,ch
		push	cx
		push	ax
		call	PrintLinePos	; E: xy in AX, number of spaces in CX
		pop	ax
		pop	cx
		pop	dx
		push	dx
		mov	ah,dh
		call	PrintLinePos	; E: xy in AX, number of spaces in CX
		pop	cx
		pop	ax
		push	ax
		push	cx
		mov	bx,offset TxUppLeftCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
PrintRect2:	mov	bx,offset TxVertLine
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
		cmp	ah,ch
		jnz	PrintRect2
		mov	bx,offset TxLwrLeftCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		pop	cx
		pop	ax
		mov	al,cl
		mov	bx,offset TxUppRghtCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
PrintRect4:	mov	bx,offset TxVertLine
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		inc	ah
		cmp	ah,ch
		jnz	PrintRect4
		mov	bx,offset TxLwrRghtCornr
		call	PrintPos	; E: xy in AX, pointer to string in BX
					; R: BX points to terminating 0
		retn
PrintRectangle	endp


;============================================================================

;		S u b r o u t i n e
; Print a line at xy position AX, length CX
; E: xy in AX, number of line-chars in CX

PrintLinePos	proc near
		mov	bx,offset TxHorizLine
		call	PrintPos
		inc	al		; increment X-position
		loop	PrintLinePos
		retn
PrintLinePos	endp


;============================================================================

;		S u b r o u t i n e
; This subroutine unused!
; Prints 3 strings on same line at X-positions AL, 39 and 72
; E: xy in AX, pointer to 3 strings in BX

	; unused code >>
Print3Str	proc near
		call	PrintPos
		mov	al,39		; X-position for second string
		call	PrintPos
		mov	al,72		; X-position for third string
		call	PrintPos
		retn
Print3Str	endp
	; << unused code


;============================================================================

;		S u b r o u t i n e
; E: xy upper left corner in AX, xy lower right corner in CX
; R:-

ClearScreenArea	proc near
		push	ax
		push	cx
		inc	al
		sub	cl,al		; width of rectangle to clear
		xor	ch,ch
		mov	bx,offset TxSpaceLine+82
		sub	bx,cx		; BX is pointer to space-char
		pop	cx		; string of required length
		pop	ax
		push	ax
		inc	ah
		sub	ch,ah		; heigh of rectangle to clear
		xchg	ch,cl
		xor	ch,ch
		add	cx,2		; CX is number of lines to clear
		pop	ax		; position of upper left corner
ClrScrArea2:	push	bx
		call	PrintPos
		add	ax,100h		; incrment Y-coord
		pop	bx
		loop	ClrScrArea2
		retn
ClearScreenArea	endp


;============================================================================

;		S u b r o u t i n e
; E:-
; R: AH = scan code, AL = character

GetKeyb		proc near
		push	si
		push	di
		push	bx
		mov	ah,0		; READ CHAR FROM BUFFER, WAIT IF EMPTY
		int	16h		; KEYBOARD
					; Return: AH = scan code, AL = character
		pop	bx
		pop	di
		pop	si
		retn
GetKeyb		endp


;============================================================================

;		S u b r o u t i n e
; E: xy in AX

PrintDot	proc near
		push	bx
		mov	bx,offset TxDot
		call	PrintPos
		pop	bx
		retn
PrintDot	endp

;============================================================================

;		S u b r o u t i n e
; E: character attribute in AL (ES must be CS)
; V:- / R:-

SetCharAttr	proc near
		mov	es:CharAttr,al
		retn
SetCharAttr	endp


 IF VL20
		db 2 dup(0)		; this should be "align 8", but MASM
					; fills with garbage instead of 0
 ELSE
		db 7 dup(0)
 ENDIF

;============================================================================

;		S u b r o u t i n e
; Swap BIOS shadow RAM image with BIOS EPROM image
; E: AL=1 to map EPROM image, =4 to map shadow RAM image
;      =2 to switch on programming voltage, =3 to switch off prog.voltage
; V: ES, FS
; R:-
		.386p
ProgChipSet	proc near
		push	ds
		pushad
		push	ax
		call	BuildGdt
		mov	bx,ds
		mov	ax,seg seg001
		mov	ds,ax
		assume ds:seg001
		mov	bp,sp
		mov	OldDS,bx		; save DS (never used)
		pop	ax
		call	ProtectedMode
		cmp	al,1
		jnz	ProgChipS2
		; map EPROM image
		mov	esi,0FF004010h		; NxVL chip
		mov	eax,fs:[esi]
		and	eax,0FFFFFFh
		mov	fs:[esi],eax
		mov	esi,0FF004014h
		mov	eax,fs:[esi]
		and	eax,0FFFFFFh
		mov	fs:[esi],eax
		jmp	short ProgChipS8

ProgChipS2:	cmp	al,4
		jnz	ProgChipS4
		; map shadow RAM image
		mov	esi,0FF004010h		; NxVL chip
		mov	eax,fs:[esi]
		or	eax,0FF000000h
		mov	fs:[esi],eax
		mov	esi,0FF004014h
		mov	eax,fs:[esi]
		or	eax,0FF000000h
		mov	fs:[esi],eax
		jmp	short ProgChipS8

ProgChipS4:	cmp	al,2
		jnz	ProgChipS6
		; enable flash write
		mov	esi,0FF00401Ch
		mov	eax,fs:[esi]
		or	eax,18000000h
		mov	fs:[esi],eax
		jmp	short ProgChipS8

ProgChipS6:	cmp	al,3
		jnz	ProgChipS8
		; disable flash write
		mov	esi,0FF00401Ch
		mov	eax,fs:[esi]
		and	eax,0E7FFFFFFh
		mov	fs:[esi],eax

ProgChipS8:	mov	esi,0FF004010h
		mov	eax,fs:[esi]
		mov	fs:[esi],eax
		call	RealMode
		mov	ax,0F000h
		mov	ds,ax
		assume ds:nothing
		mov	ax,ds:0
		popad
		pop	ds
		assume ds:nothing
		retn
ProgChipSet	endp


; ===========================================================================

;		S u b r	o u t i	n e
; sets up GDT and loads GDTR
; E:-, V:-, R:-

BuildGdt	proc near
		pushad
		push	ds
		mov	ax,seg seg001
		mov	ds,ax
		assume ds:seg001
		; calc code segment base address into linear address
		; and insert into GDT[1] for protected mode code selector
		xor	eax,eax
		mov	ax,cs
		shl	eax,4
		mov	bx,offset GDTable
		mov	[bx+0Ah],ax	; lowword base
		shr	eax,10h
		mov	[bx+0Ch],al	; highbyte base
		; calc seg001 base address into linear address
		; and insert into GDT[2] for protected mode data selector
		xor	eax,eax
		mov	ax,seg seg001
		shl	eax,4
		push	eax
		mov	[bx+12h],ax
		shr	eax,10h
		mov	[bx+14h],al
		; insert linear address of GDTable into struc for LGDT op
		pop	eax
		add	eax,offset GDTable
		mov	dword ptr DescTableReg+2,eax ; base address of GDT
		lgdt	fword ptr DescTableReg
		pop	ds
		assume ds:nothing
		popad
		retn
BuildGdt	endp


; ===========================================================================

;		S u b r	o u t i	n e
; turns on protected mode
; E:-, V:-
; R: CS = 8 (flashv selector), DS = 10h (seg001 selector),
;    FS = 18h (flat data selector)

ProtectedMode	proc near
		pushad
		call	SaveInts
		mov	al,2
		out	92h,al		; A20 active
		mov	eax,cr0
		or	eax,1		; Protection enable
		mov	cr0,eax
FarJumpProt:	db	0EAh		; FAR_JMP $+4,8
		dw	$+4		; flush queue and load CS selector (flashv)
		dw	8		; and cache
		mov	ax,10h		; index for DS-descriptor (seg001)
		mov	ds,ax		; load DS selector and cache
		mov	ax,18h		; index for FS-descriptor (flat data)
		mov	fs,ax		; load FS selector and cache
		popad
		retn
ProtectedMode	endp


; ===========================================================================

;		S u b r	o u t i	n e
; E:-, V:-, R:-

SaveInts	proc near
		push	ax
		in	al,21h		; Interrupt controller,	8259A.
		xchg	al,ah
		mov	al,0FFh
		out	21h,al		; Interrupt controller,	8259A.
		jmp	short $+2	; IODelay
		in	al,0A1h		; Interrupt Controller #2, 8259A
		mov	ds:IntMask,ax
		mov	al,0FFh
		out	0Ah,al		; DMA controller, 8237A-5.
					; single mask bit register
					; 0-1: select channel (00=0; 01=1; 10=2; 11=3)
					; 2: 1=set mask	for channel; 0=clear mask (enable)
		jmp	short $+2	; IODelay
		pop	ax
		retn
SaveInts	endp


; ===========================================================================

;		S u b r	o u t i	n e
; turns off protected mode
; E:-, V:-
; R: CS = flashv segment, DS = ES = seg001 segment

RealMode	proc near
		pushad
		mov	ax,10h		; selector for seg001
		mov	es,ax
		mov	fs,ax
		mov	eax,0		; Protection enable aus
		mov	cr0,eax
FarJumpReal:	jmp	far ptr	RealMode2 ; flush queue and load CS
RealMode2:	mov	ax,seg seg001
		mov	ds,ax
		assume ds:seg001
		mov	es,ax
		assume es:seg001
		call	RestoreInts
		popad
		retn
RealMode	endp


; ===========================================================================

;		S u b r	o u t i	n e
; E:-, V:-, R:-

RestoreInts	proc near
		push	ax
		mov	ax,IntMask
		out	0A1h,al		; Interrupt Controller #2, 8259A
		jmp	short $+2	; IODelay
		xchg	al,ah
		out	21h,al		; Interrupt controller,	8259A.
		jmp	short $+2	; IODelay
		pop	ax
		retn
RestoreInts	endp

flashv		ends


; Segment type:	Regular
seg001		segment	byte public 'DATA' use16
		assume cs:seg001
		;org 0Bh
		assume es:nothing, ss:nothing, ds:nothing, fs:nothing, gs:nothing
		db    0
IntMask		dw 0
OldDS		dw 0			; saved DS (never used)
		db    0
		db    0
		db    0
		db    0
		db    0
		db    0
		db    0
		db    0
		db    0
		db    0

DescTableReg	dw 5*8h	; limit of GDT: 5 entries
		dd 0	; base address <-- inserted at runtime

		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db 0DEh	;
		db  9Bh	;
		db  57h	;
		db  13h	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		db    0	;
		dw 2000h
		db '         ',0

GDTable:	; GDT[0]: unused
		db 8 dup(0)

		; GDT[1]: code 64K from flashv
		dw 0FFFFh	; limit (lower 16bit)
		db 3 dup(0)	; base (lower 24bit) <-- inserted at runtime
		db 10011010b	; 9Ah flags: present, privilege level 0, execute/read
		db 0		; limit (upper 4bit)
		db 0		; base (upper 8bit)

		; GDT[2]: data 64K from seg001
		dw 0FFFFh	; limit (lower 16bit)
		db 3 dup(0)	; base (lower 24bit) <-- inserted at runtime
		db 10010011b	; 93h flags: present, privilege level 0,
				; read/write, accessed
		db 0		; limit (upper 4bit)
		db 0		; base (upper 8bit)

		; GDT[3]: data 4GB from 0 (flat data)
		dw 0FFFFh	; limit (lower 16bit)
		db 3 dup(0)	; base (lower 24bit)
		db 10010011b	; 93h flags: present, privilege level 0,
				; read/write, accessed
		db 9Fh		; limit (upper 4bit), granularity=4K
		db 0		; base (upper 8bit)

		; GDT[4]: data 64K from 0B8000h (video memory)
		dw 0FFFFh	; limit (lower 16bit)
		db 0,80h,0Bh	; base (lower 24bit)
		db 10010011b	; 93h flags: present, privilege level 0,
				; read/write, accessed
		db 0		; limit (upper 4bit)
		db 0		; base (upper 8bit)
seg001		ends


		end Start