/*
 * cpu_id.c - let windows applications see a 486 processor
 *
 * Enables hypercode emulation of i486 application instructions in
 * the NexGen Nx586 and forces WF_CPU486 in GetWinFlags result.
 *
 * Created by NexGen Inc.
 * re-sourced by Herbert Oppmann (herby@memotech.franken.de)
 */

//--- INCLUDES ---

#include <windows.h>
#include "cpu_id.h"
#include "prochook.h"
#include "goodies.h"

//--- GLOBALS ---

// Global stuff
HANDLE hInst;
HWND   ghWnd;
// Magic cookie for GetProcAddress hook
NPHOOKCHILD npHookChild;

/***************************************************************************
 * WinMain() - Main Windows routine
 ***************************************************************************/

int PASCAL WinMain(						//- ENTRY
	HANDLE hInstance,
	HANDLE hPrevInstance,
	LPSTR lpszCmdLine,
	int nCmdShow
){										//- LOCAL
	static char	szAppName[]="CPU_ID";
    WNDCLASS	wndclass;
    MSG			msg;
    HACCEL		hAccTable;
	HWND		hWnd;

    hInst = hInstance;
    if(!hPrevInstance)
    {
        wndclass.style         = CS_HREDRAW | CS_VREDRAW;
        wndclass.lpfnWndProc   = WndProc;
        wndclass.cbClsExtra    = 0;
        wndclass.cbWndExtra    = DLGWINDOWEXTRA;
        wndclass.hInstance     = hInstance;
        wndclass.hIcon         = LoadIcon(hInstance,szAppName);
        wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wndclass.hbrBackground = COLOR_WINDOW + 1;
        wndclass.lpszMenuName  = szAppName;
        wndclass.lpszClassName = szAppName;

        if(!RegisterClass(&wndclass))
            return -1;

	    if((hWnd=CreateDialog(hInstance,szAppName,0,NULL)) == NULL)
	        return -1;
        // Make HWND global
		ghWnd = hWnd;
	    ShowWindow(hWnd,nCmdShow);
	
		hAccTable = LoadAccelerators(hInstance,szAppName);
	
	    while(GetMessage(&msg,NULL,0,0))
	    {
			if (!TranslateAccelerator(hWnd, hAccTable, &msg)) {
	        	if((!IsWindow(hWnd)) ||
		           (!IsDialogMessage(hWnd,&msg)))
	    	    {
	        	    TranslateMessage(&msg);
	            	DispatchMessage(&msg);
		        }
	    	}
	    }
	    return msg.wParam;
    } else {
		hWnd = FindWindow(szAppName, NULL);
		UpdateWindow(hWnd);
		ShowWindow(hWnd, 13); // SW_xx?
		SetActiveWindow(hWnd);
		return 0;
    }
}

/***************************************************************************
 *  WndProc()
 *
 *  Message handler for the application window.
 *
 *  Returns long - Variable, depends on message.
 ***************************************************************************/
          
#pragma optimize("e", off)
long FAR PASCAL __export WndProc(		//- ENTRY
	HWND hWnd,
	UINT message,
	WPARAM wParam,
	LPARAM lParam
){										//- LOCAL
    FARPROC lpfnNewGetWF;
    static FARPROC lpAboutDlg;
    static HANDLE hCreateInst;

    switch (message)
    {
        case WM_CREATE:
            // Center the window on the screen
            CenterWindow(hWnd);
            
            // Set up the hook to our new function
            lpfnNewGetWF = MakeProcInstance((FARPROC)NewGetWinFlags, hInst);
			npHookChild = SetProcAddress((FARPROC)GetWinFlags, lpfnNewGetWF, FALSE);
			hCreateInst = ((CREATESTRUCT FAR *)lParam)->hInstance;
			lpAboutDlg = MakeProcInstance(AboutDlgProc, hCreateInst);
			PostMessage(hWnd,WM_486MODE,0,0L);
            break;
        
        case WM_ENDSESSION:
        case WM_DESTROY:
			// disable hypercode emulation of i486 application instructions
        	_asm {
        		mov si, 0B00Eh
        		mov al, 4
        		mov ah, 4Ch
        		nxop
        	}
            // Delete the reference to the hook
        	SetProcRelease(npHookChild);
            PostQuitMessage(0);
            break;

		case WM_486MODE:
			SetFocus(GetDlgItem(hWnd, IDM_NOTES));
			// enable hypercode emulation of i486 application instructions
        	_asm {
        		mov si, 0B00Eh
        		mov al, 3
        		mov ah, 4Ch
        		nxop
        	}
			break;

        // Processing for the buttons...
        case WM_COMMAND:
            switch(wParam)
            {
                case IDM_NOTES:
					DialogBox(hCreateInst, "NotesBox", hWnd, lpAboutDlg);
    				return NULL;
                    break;
                    
                case IDM_TERMS:
					DialogBox(hCreateInst, "TermsBox", hWnd, lpAboutDlg);
    				return NULL;
                    break;

                case IDM_ABOUT:
					DialogBox(hCreateInst, "AboutBox", hWnd, lpAboutDlg);
				    return NULL;
					break;

                case IDB_QUIT:
					// disable hypercode emulation of i486 application instructions
		        	_asm {
		        		mov si, 0B00Eh
		        		mov al, 4
		        		mov ah, 4Ch
		        		nxop
		        	}
                    PostMessage(hWnd,WM_CLOSE,0,0L);
                    break;
            }
            break;
            
        default:
            return DefWindowProc(hWnd, message, wParam, lParam);
            break;
    }
    return NULL;
}
#pragma optimize("e", on)

/***************************************************************************
 * AboutDlgProc()
 *
 * Dialog procedure for Notes-, Terms- and AboutBox.
 *
 *	returns BOOL - TRUE if processed, FALSE otherwise
 **************************************************************************/

BOOL FAR PASCAL __export AboutDlgProc(	//- ENTRY
	HWND hDlg,
	UINT message,
	WPARAM wParam,
	LPARAM lParam
){										//- LOCAL
    switch(message)
    {
        case WM_COMMAND:
            switch (wParam){
                case IDOK:
                case IDCANCEL:
                    EndDialog(hDlg, FALSE);
                    return TRUE;
                    break;
            }
            break;

		case WM_INITDIALOG:
			return TRUE;
			break;
	}
	return FALSE;
}

/***************************************************************************
 * NewGetWinFlag()
 *
 * Callback for hooked function.
 *
 *	returns DWORD - see GetWinFlags()
 **************************************************************************/

DWORD WINAPI __export NewGetWinFlags()
{
	DWORD dwWinFlags;
    static char szItem[103]; // unused
    HTASK       wCurTask;    // leftover from GPA.C, unused

    // Unhook our reference
	ProcUnhook(npHookChild);
    // Reissue the call to GetWinFlags 
	dwWinFlags = GetWinFlags();
    // Rehook our reference
	ProcHook(npHookChild);

    // Get the caller's task handle
	wCurTask = GetCurrentTask(); // leftover from GPA.C, unused

    // Return the modified value returned from the call to GetWinFlags
	return (dwWinFlags & ~(WF_CPU286|WF_CPU386)) | WF_CPU486;
}
